"""
The Arcade Library

A Python simple, easy to use module for creating 2D games.
"""

# flake8: noqa: E402
# Error out if we import Arcade with an incompatible version of Python.
import sys
import os
from typing import Final

from pathlib import Path

if sys.version_info[0] < 3 or (sys.version_info[0] == 3 and sys.version_info[1] < 10):
    sys.exit("The Arcade Library requires Python 3.10 or higher.")


def configure_logging(level: int | None = None):
    """Set up basic logging.

    Args:
        level: The log level. Defaults to DEBUG.
    """
    import logging

    level = level or logging.DEBUG
    LOG = logging.getLogger(__name__)
    # Do not add a new handler if we already have one
    if not LOG.handlers:
        LOG.propagate = False
        LOG.setLevel(level)
        ch = logging.StreamHandler()
        ch.setLevel(level)
        ch.setFormatter(
            logging.Formatter("%(relativeCreated)s %(name)s %(levelname)s - %(message)s")
        )
        LOG.addHandler(ch)


# The following is used to load ffmpeg libraries.
# Currently Arcade is only shipping binaries for Mac OS
# as ffmpeg is not needed for support on Windows and Linux.
# However it is setup to load ffmpeg if the binaries are present
# on Windows and Linux. So if you need ffmpeg you can simply
# drop the binaries in the "lib" folder of Arcade
lib_location = Path(__file__).parent.absolute()
lib_location = lib_location / "lib"

if sys.platform == "darwin" or sys.platform.startswith("linux"):
    if "LD_LIBRARY_PATH" in os.environ:
        os.environ["LD_LIBRARY_PATH"] += ":" + str(lib_location)
    else:
        os.environ["LD_LIBRARY_PATH"] = str(lib_location)
else:
    os.environ["PATH"] += str(lib_location)

import pyglet

# Enable HiDPI support using stretch mode
if os.environ.get("ARCADE_TEST"):
    pyglet.options.dpi_scaling = "real"
else:
    pyglet.options.dpi_scaling = "stretch"

# Env variable shortcut for headless mode
headless: Final[bool] = bool(os.environ.get("ARCADE_HEADLESS"))
if headless:
    pyglet.options.headless = headless


# from arcade import utils
# Disable shadow window on macs and in headless mode.
# if sys.platform == "darwin" or os.environ.get("ARCADE_HEADLESS") or utils.is_raspberry_pi():
# NOTE: We always disable shadow window now to have consistent behavior across platforms.
pyglet.options.shadow_window = False

# Imports from modules that don't do anything circular

# Complex imports with potential circularity
from .window_commands import close_window
from .window_commands import exit
from .window_commands import finish_render
from .window_commands import get_display_size
from .window_commands import get_window
from .window_commands import schedule
from .window_commands import run
from .window_commands import set_background_color
from .window_commands import set_window
from .window_commands import start_render
from .window_commands import unschedule
from .window_commands import schedule_once

from .sections import Section, SectionManager

from .application import MOUSE_BUTTON_LEFT
from .application import MOUSE_BUTTON_MIDDLE
from .application import MOUSE_BUTTON_RIGHT
from .application import NoOpenGLException
from .application import View
from .application import Window
from .application import get_screens
from .application import open_window

from .texture import Texture
from .texture import TextureCacheManager
from .texture import SpriteSheet
from .texture import load_spritesheet
from .texture import load_texture
from .texture import load_image
from .texture import make_circle_texture
from .texture import make_soft_circle_texture
from .texture import make_soft_square_texture
from .texture import get_default_image
from .texture import get_default_texture

from .draw import get_points_for_thick_line
from .draw import draw_arc_filled
from .draw import draw_arc_outline
from .draw import draw_circle_filled
from .draw import draw_circle_outline
from .draw import draw_ellipse_filled
from .draw import draw_ellipse_outline
from .draw import draw_line
from .draw import draw_line_strip
from .draw import draw_lines
from .draw import draw_lrbt_rectangle_filled
from .draw import draw_lrbt_rectangle_outline
from .draw import draw_parabola_filled
from .draw import draw_parabola_outline
from .draw import draw_point
from .draw import draw_points
from .draw import draw_polygon_filled
from .draw import draw_polygon_outline
from .draw import draw_rect_filled
from .draw import draw_rect_outline
from .draw import draw_texture_rect
from .draw import draw_sprite
from .draw import draw_sprite_rect
from .draw import draw_triangle_filled
from .draw import draw_triangle_outline
from .draw import draw_lbwh_rectangle_filled
from .draw import draw_lbwh_rectangle_outline
from .screenshot import get_image
from .screenshot import get_pixel

# We don't have joysticks game controllers in headless mode
if not headless:
    from .joysticks import get_game_controllers
    from .joysticks import get_joysticks
    from .controller import ControllerManager
    from .controller import get_controllers

from .sound import Sound
from .sound import load_sound
from .sound import play_sound
from .sound import stop_sound

from .sprite import FACE_DOWN
from .sprite import FACE_LEFT
from .sprite import FACE_RIGHT
from .sprite import FACE_UP
from .sprite import TextureAnimationSprite
from .sprite import load_animated_gif
from .sprite import AnimatedWalkingSprite
from .sprite import TextureAnimation
from .sprite import TextureKeyframe
from .sprite import PyMunk
from .sprite import PymunkMixin
from .sprite import SpriteType
from .sprite import SpriteType_co
from .sprite import Sprite
from .sprite import BasicSprite

# from .sprite import SimpleSprite
from .sprite import SpriteCircle
from .sprite import SpriteSolidColor

from .sprite_list import SpriteList
from .sprite_list import SpriteSequence
from .sprite_list import check_for_collision
from .sprite_list import check_for_collision_with_list
from .sprite_list import check_for_collision_with_lists
from .sprite_list import get_closest_sprite
from .sprite_list import get_sprites_at_exact_point
from .sprite_list import get_sprites_at_point
from .sprite_list import get_distance_between_sprites
from .sprite_list import get_sprites_in_rect

from .sprite_list import SpatialHash

from .scene import Scene
from .scene import SceneKeyError

from .physics_engines import PhysicsEnginePlatformer
from .physics_engines import PhysicsEngineSimple

from .tilemap import load_tilemap
from .tilemap import TileMap

from .pymunk_physics_engine import PymunkPhysicsEngine
from .pymunk_physics_engine import PymunkPhysicsObject
from .pymunk_physics_engine import PymunkException

from .version import VERSION

from .paths import has_line_of_sight
from .paths import AStarBarrierList
from .paths import astar_calculate_path

from .context import ArcadeContext

from .texture_atlas import DefaultTextureAtlas

from .perf_info import enable_timings
from .perf_info import print_timings
from .perf_info import get_fps
from .perf_info import get_timings
from .perf_info import clear_timings
from .perf_info import timings_enabled
from .perf_info import disable_timings

from .perf_graph import PerfGraph

from .camera import Camera2D

from .types.rect import Rect, LRBT, LBWH, XYWH

# Module imports
from arcade import color as color
from arcade import csscolor as csscolor
from arcade import uicolor as uicolor
from arcade import camera as camera
from arcade import key as key
from arcade import resources as resources
from arcade import types as types
from arcade import math as math
from arcade import shape_list as shape_list
from arcade import hitbox as hitbox
from arcade import experimental as experimental
from arcade.types import rect

# For ease of access for beginners
from pyglet.math import Vec2, Vec3, Vec4

from .text import (
    draw_text,
    load_font,
    create_text_sprite,
    Text,
)

__all__ = [
    "AStarBarrierList",
    "AnimatedWalkingSprite",
    "TextureAnimationSprite",
    "TextureAnimation",
    "TextureKeyframe",
    "ArcadeContext",
    "ControllerManager",
    "FACE_DOWN",
    "FACE_LEFT",
    "FACE_RIGHT",
    "FACE_UP",
    "MOUSE_BUTTON_LEFT",
    "MOUSE_BUTTON_MIDDLE",
    "MOUSE_BUTTON_RIGHT",
    "NoOpenGLException",
    "PerfGraph",
    "PhysicsEnginePlatformer",
    "PhysicsEngineSimple",
    "PyMunk",
    "PymunkException",
    "PymunkPhysicsEngine",
    "PymunkPhysicsObject",
    "Rect",
    "LBWH",
    "LRBT",
    "XYWH",
    "Section",
    "SectionManager",
    "Scene",
    "SceneKeyError",
    "Sound",
    "BasicSprite",
    "Sprite",
    "SpriteType",
    "SpriteType_co",
    "PymunkMixin",
    "SpriteCircle",
    "SpriteList",
    "SpriteSequence",
    "SpriteSolidColor",
    "Text",
    "Texture",
    "TextureCacheManager",
    "SpriteSheet",
    "DefaultTextureAtlas",
    "TileMap",
    "VERSION",
    "Vec2",
    "Vec3",
    "Vec4",
    "View",
    "Window",
    "astar_calculate_path",
    "check_for_collision",
    "check_for_collision_with_list",
    "check_for_collision_with_lists",
    "close_window",
    "disable_timings",
    "draw_arc_filled",
    "draw_arc_outline",
    "draw_circle_filled",
    "draw_circle_outline",
    "draw_ellipse_filled",
    "draw_ellipse_outline",
    "draw_line",
    "draw_line_strip",
    "draw_lines",
    "draw_lrbt_rectangle_filled",
    "draw_lrbt_rectangle_filled",
    "draw_lrbt_rectangle_outline",
    "draw_lrbt_rectangle_outline",
    "draw_parabola_filled",
    "draw_parabola_outline",
    "draw_point",
    "draw_points",
    "draw_polygon_filled",
    "draw_polygon_outline",
    "draw_rect_filled",
    "draw_rect_outline",
    "draw_text",
    "draw_texture_rect",
    "draw_sprite",
    "draw_sprite_rect",
    "draw_triangle_filled",
    "draw_triangle_outline",
    "draw_lbwh_rectangle_filled",
    "draw_lbwh_rectangle_outline",
    "enable_timings",
    "exit",
    "finish_render",
    "get_closest_sprite",
    "get_display_size",
    "get_distance_between_sprites",
    "get_sprites_in_rect",
    "get_controllers",
    "get_game_controllers",
    "get_image",
    "get_joysticks",
    "get_pixel",
    "get_points_for_thick_line",
    "get_screens",
    "get_sprites_at_exact_point",
    "get_sprites_at_point",
    "SpatialHash",
    "get_timings",
    "create_text_sprite",
    "clear_timings",
    "get_window",
    "get_fps",
    "has_line_of_sight",
    "load_animated_gif",
    "load_font",
    "load_sound",
    "load_spritesheet",
    "load_texture",
    "load_image",
    "make_circle_texture",
    "make_soft_circle_texture",
    "make_soft_square_texture",
    "open_window",
    "print_timings",
    "play_sound",
    "load_tilemap",
    "run",
    "schedule",
    "set_background_color",
    "set_window",
    "start_render",
    "stop_sound",
    "timings_enabled",
    "unschedule",
    "schedule_once",
    "get_default_texture",
    "get_default_image",
    "hitbox",
    "experimental",
    "rect",
    "color",
    "csscolor",
    "uicolor",
    "key",
    "resources",
    "types",
    "math",
    "shape_list",
    "Camera2D",
]

__version__ = VERSION

# Piggyback on pyglet's doc run detection
if not getattr(sys, "is_pyglet_doc_run", False):
    # Load additional game controller mappings to Pyglet
    if not headless:
        try:
            import pyglet.input.controller

            mappings_file = resources.resolve(":system:gamecontrollerdb.txt")
            # TODO: remove string conversion once fixed upstream
            pyglet.input.controller.add_mappings_from_file(str(mappings_file))
        except AssertionError:
            pass



from arcade.camera.grips.strafe import strafe
from arcade.camera.grips.rotate import (
    rotate_around_forward,
    rotate_around_up,
    rotate_around_right,
)
from arcade.camera.grips.screen_shake_2d import ScreenShake2D
from arcade.camera.grips.constrain import (
    constrain_x,
    constrain_y,
    constrain_z,
    constrain_xy,
    constrain_yz,
    constrain_xz,
    constrain_xyz,
    constrain_boundary_x,
    constrain_boundary_y,
    constrain_boundary_z,
    constrain_boundary_xy,
    constrain_boundary_yz,
    constrain_boundary_xz,
    constrain_boundary_xyz,
)
from arcade.camera.grips.position import look_at, orbit


__all__ = (
    "strafe",
    "rotate_around_right",
    "rotate_around_up",
    "rotate_around_forward",
    "ScreenShake2D",
    "constrain_x",
    "constrain_y",
    "constrain_z",
    "constrain_xy",
    "constrain_yz",
    "constrain_xz",
    "constrain_xyz",
    "constrain_boundary_x",
    "constrain_boundary_y",
    "constrain_boundary_z",
    "constrain_boundary_xy",
    "constrain_boundary_yz",
    "constrain_boundary_xz",
    "constrain_boundary_xyz",
    "look_at",
    "orbit",
)



from math import pi, tan

from pyglet.math import Mat4, Vec2, Vec3, Vec4

from arcade.camera.data_types import (
    CameraData,
    OrthographicProjectionData,
    PerspectiveProjectionData,
)
from arcade.types import Point


def generate_view_matrix(camera_data: CameraData) -> Mat4:
    """
    Using the ViewData it generates a view matrix from the pyglet Mat4 look at function
    """
    # Even if forward and up are normalized floating point error means every vector
    # must be normalized.
    fo = Vec3(*camera_data.forward).normalize()  # Forward Vector
    up = Vec3(
        *camera_data.up
    )  # Initial Up Vector (Not necessarily perpendicular to forward vector)
    ri = fo.cross(up).normalize()  # Right Vector
    up = ri.cross(fo).normalize()  # Up Vector
    po = Vec3(*camera_data.position)

    # fmt: off
    return Mat4(
        ri.x, up.x, -fo.x, 0.0,
        ri.y, up.y, -fo.y, 0.0,
        ri.z, up.z, -fo.z, 0.0,
        -ri.dot(po), -up.dot(po), fo.dot(po), 1.0
    )
    # fmt: on


def generate_orthographic_matrix(
    perspective_data: OrthographicProjectionData, zoom: float = 1.0
) -> Mat4:
    """
    Using the OrthographicProjectionData a projection matrix is generated where
    the size of an object is not affected by depth.

    Generally keep the scale value to integers or negative powers of integers
    (``2^-1, 3^-1, 2^-2``, etc.) to keep the pixels uniform in size. Avoid a zoom of 0.0.
    """

    # Scale the projection by the zoom value. Both the width and the height
    # share a zoom value to avoid ugly stretching.
    left = perspective_data.left / zoom
    right = perspective_data.right / zoom
    bottom = perspective_data.bottom / zoom
    top = perspective_data.top / zoom

    z_near, z_far = perspective_data.near, perspective_data.far

    width = right - left
    height = top - bottom
    depth = z_far - z_near

    sx = 2.0 / width
    sy = 2.0 / height
    sz = 2.0 / -depth

    tx = -(right + left) / width
    ty = -(top + bottom) / height
    tz = -(z_far + z_near) / depth

    # fmt: off
    return Mat4(
        sx, 0.0, 0.0, 0.0,
        0.0,  sy, 0.0, 0.0,
        0.0, 0.0,  sz, 0.0,
        tx,  ty,  tz, 1.0
    )
    # fmt: on


def generate_perspective_matrix(
    perspective_data: PerspectiveProjectionData, zoom: float = 1.0
) -> Mat4:
    """
    Using the OrthographicProjectionData a projection matrix is generated where
    the size of the objects is not affected by depth.

    Generally keep the scale value to integers or negative powers of integers
    (``2^-1, 3^-1, 2^-2``, etc.) to keep the pixels uniform in size. Avoid a zoom of 0.0.
    """
    fov = perspective_data.fov / zoom
    z_near, z_far, aspect = (
        perspective_data.near,
        perspective_data.far,
        perspective_data.aspect,
    )

    xy_max = z_near * tan(fov * pi / 360)
    y_min = -xy_max
    x_min = -xy_max

    width = xy_max - x_min
    height = xy_max - y_min
    depth = z_far - z_near
    q = -(z_far + z_near) / depth
    qn = -2 * z_far * z_near / depth

    w = 2 * z_near / width
    w = w / aspect
    h = 2 * z_near / height

    # fmt: off
    return Mat4(
        w, 0, 0, 0,
        0, h, 0, 0,
        0, 0, q, -1,
        0, 0, qn, 0
    )
    # fmt: on


def project_orthographic(
    world_coordinate: Point,
    viewport: tuple[int, int, int, int],
    view_matrix: Mat4,
    projection_matrix: Mat4,
) -> Vec2:
    """
    Project a world coordinate to a screen coordinate using an orthographic projection.

    Args:
        world_coordinate: The world coordinate to project.
        viewport: The viewport of the camera.
        view_matrix: The view matrix of the camera.
        projection_matrix: The projection matrix of the camera.
    """
    x, y, *_z = world_coordinate
    z = 0.0 if not _z else _z[0]

    world_position = Vec4(x, y, z, 1.0)

    projected_position = projection_matrix @ view_matrix @ world_position

    screen_coordinate_x = viewport[0] + (0.5 * projected_position.x + 0.5) * viewport[2]
    screen_coordinate_y = viewport[1] + (0.5 * projected_position.y + 0.5) * viewport[3]

    return Vec2(screen_coordinate_x, screen_coordinate_y)


def unproject_orthographic(
    screen_coordinate: Point,
    viewport: tuple[int, int, int, int],
    view_matrix: Mat4,
    projection_matrix: Mat4,
) -> Vec3:
    """
    Unproject a screen coordinate to a world coordinate using an orthographic projection.

    Args:
        screen_coordinate: The screen coordinate to unproject.
        viewport: The viewport of the camera.
        view_matrix: The view matrix of the camera.
        projection_matrix: The projection matrix of the camera.
    """
    x, y, *_z = screen_coordinate
    z = 0.0 if not _z else _z[0]

    screen_x = 2.0 * (x - viewport[0]) / viewport[2] - 1
    screen_y = 2.0 * (y - viewport[1]) / viewport[3] - 1

    _projection = ~projection_matrix
    _view = ~view_matrix

    _unprojected_position = _projection @ Vec4(screen_x, screen_y, 0.0, 1.0)
    _world_position = _view @ Vec4(_unprojected_position.x, _unprojected_position.y, z, 1.0)

    return Vec3(_world_position.x, _world_position.y, _world_position.z)


def project_perspective(
    world_coordinate: Point,
    viewport: tuple[int, int, int, int],
    view_matrix: Mat4,
    projection_matrix: Mat4,
) -> Vec2:
    """
    Project a world coordinate to a screen coordinate using a perspective projection.

    Args:
        world_coordinate: The world coordinate to project.
        viewport: The viewport of the camera.
        view_matrix: The view matrix of the camera.
        projection_matrix: The projection matrix of the camera.
    """
    x, y, *_z = world_coordinate
    z = 1.0 if not _z else _z[0]

    world_position = Vec4(x, y, z, 1.0)

    semi_projected_position = projection_matrix @ view_matrix @ world_position
    div_val = semi_projected_position.w

    projected_x = semi_projected_position.x / div_val
    projected_y = semi_projected_position.y / div_val

    screen_coordinate_x = viewport[0] + (0.5 * projected_x + 0.5) * viewport[2]
    screen_coordinate_y = viewport[1] + (0.5 * projected_y + 0.5) * viewport[3]

    return Vec2(screen_coordinate_x, screen_coordinate_y)


def unproject_perspective(
    screen_coordinate: Point,
    viewport: tuple[int, int, int, int],
    view_matrix: Mat4,
    projection_matrix: Mat4,
) -> Vec3:
    """
    Unproject a screen coordinate to a world coordinate using a perspective projection.

    Args:
        screen_coordinate: The screen coordinate to unproject.
        viewport: The viewport of the camera.
        view_matrix: The view matrix of the camera.
        projection_matrix: The projection matrix of
    """
    x, y, *_z = screen_coordinate
    z = 1.0 if not _z else _z[0]

    screen_x = 2.0 * (x - viewport[0]) / viewport[2] - 1
    screen_y = 2.0 * (y - viewport[1]) / viewport[3] - 1

    screen_x *= z
    screen_y *= z

    projected_position = Vec4(screen_x, screen_y, 1.0, 1.0)

    view_position = ~projection_matrix @ projected_position
    world_position = ~view_matrix @ Vec4(view_position.x, view_position.y, z, 1.0)

    return Vec3(world_position.x, world_position.y, world_position.z)



"""
Sprite Rotation Around a Point, With A Tank

Games often include elements that rotate toward targets. Common
examples include gun turrets on vehicles and towers. In 2D games,
these rotating parts are usually implemented as sprites that move
relative to whatever they're attached to.

There's a catch to this: you have to rotate these parts around their
attachment points rather than the centers of their sprites. Otherwise,
the rotation will look wrong!

To illustrate the difference, this example uses a player-controllable
tank with a barrel that follows the mouse. You can press P to switch
between two ways of rotating the barrel:
1. Correctly, with the barrel's rear against the tank's center
2. Incorrectly, around the barrel's center pinned to the tank's

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_rotate_around_tank
"""
import math
import arcade
from arcade.types import Point
from arcade.math import (
    get_angle_radians,
    rotate_point,
    get_angle_degrees,
)

TANK_SPEED_PIXELS = 64  # How many pixels per second the tank travels
TANK_TURN_SPEED_DEGREES = 70  # How fast the tank's body can turn


# This is half the length of the barrel sprite.
# We use it to ensure the barrel's rear sits in the middle of the tank
TANK_BARREL_LENGTH_HALF = 15


WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Rotating Tank Example"


# These paths are built-in resources included with arcade
TANK_BODY = ":resources:images/topdown_tanks/tankBody_dark_outline.png"
TANK_BARREL = ":resources:images/topdown_tanks/tankDark_barrel3_outline.png"


class RotatingSprite(arcade.Sprite):
    """
    Sprite subclass which can be rotated around a point.

    This version of the class always changes the angle of the sprite.
    Other games might not rotate the sprite. For example, moving
    platforms in a platformer wouldn't rotate.
    """
    def rotate_around_point(self, point: Point, degrees: float):
        """
        Rotate the sprite around a point by the set amount of degrees

        Args:
            point: The point that the sprite will rotate about
            degrees: How many degrees to rotate the sprite
        """

        # Make the sprite turn as its position is moved
        self.angle += degrees

        # Move the sprite along a circle centered around the passed point
        self.position = rotate_point(
            self.center_x, self.center_y,
            point[0], point[1], degrees)

    def face_point(self, point: Point):
        self.angle = get_angle_degrees(*self.position, *point)


class GameView(arcade.View):

    def __init__(self):
        super().__init__()

        # Set Background to be green
        self.window.background_color = arcade.csscolor.SEA_GREEN

        # The tank and barrel sprite
        self.tank = arcade.Sprite(TANK_BODY)
        self.tank.position = self.window.center

        self.barrel = RotatingSprite(TANK_BARREL)
        self.barrel.position = (
            self.tank.center_x, self.tank.center_y - TANK_BARREL_LENGTH_HALF
        )

        self.tank_direction = 0.0  # Forward & backward throttle
        self.tank_turning = 0.0  # Turning strength to the left or right

        self.mouse_pos = 0, 0

        self.tank_sprite_list = arcade.SpriteList()
        self.tank_sprite_list.extend([self.tank, self.barrel])

        self._correct = True
        self.correct_text = arcade.Text(
            "If the turret rotation is incorrect press P to reset it",
            self.window.center_x, WINDOW_HEIGHT - 25,
            anchor_x='center')

        self.control_text = arcade.Text(
            "WASD to move tank, Mouse to aim",
            self.window.center_x, 15,
            anchor_x='center')

    def on_draw(self):
        self.clear()
        self.tank_sprite_list.draw()

        self.control_text.draw()
        self.correct_text.draw()

    def on_update(self, delta_time: float):
        self.move_tank(delta_time)

    def move_tank(self, delta_time):
        """
        Perform all calculations for moving the tank's body and barrel
        """

        # Rotate the tank's body in place without changing position
        # We'll rotate the barrel after updating the entire tank's x & y
        self.tank.angle += (
            TANK_TURN_SPEED_DEGREES * self.tank_turning * delta_time
        )

        # Calculate how much the tank should move forward or back
        move_magnitude = self.tank_direction * TANK_SPEED_PIXELS * delta_time
        x_dir = math.sin(self.tank.radians) * move_magnitude
        y_dir = math.cos(self.tank.radians) * move_magnitude

        # Move the tank's body
        self.tank.position = (
            self.tank.center_x + x_dir,
            self.tank.center_y + y_dir
        )

        # Move the barrel with the body
        self.barrel.position = (
            self.barrel.center_x + x_dir,
            self.barrel.center_y + y_dir
        )

        # Begin rotating the barrel by finding the angle to the mouse
        mouse_angle = get_angle_degrees(
            self.tank.center_x, self.tank.center_y,
            self.mouse_pos[0], self.mouse_pos[1])

        # Compensate for fact that the barrel sits horzontally rather than virtically
        mouse_angle -= 90

        # Rotate the barrel sprite with one end at the tank's center
        # Subtract the old angle to get the change in angle
        angle_change = mouse_angle - self.barrel.angle

        self.barrel.rotate_around_point(self.tank.position, angle_change)

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.W:
            self.tank_direction += 1
        elif symbol == arcade.key.S:
            self.tank_direction -= 1
        elif symbol == arcade.key.A:
            self.tank_turning -= 1
        elif symbol == arcade.key.D:
            self.tank_turning += 1
        elif symbol == arcade.key.P:
            angle = self.barrel.angle
            self.barrel.angle = 0
            self.barrel.position = (
                self.tank.center_x,
                self.tank.center_y - TANK_BARREL_LENGTH_HALF
            )
            self.barrel.rotate_around_point(self.tank.position, angle)

    def on_key_release(self, symbol: int, modifiers: int):
        if symbol == arcade.key.W:
            self.tank_direction -= 1
        elif symbol == arcade.key.S:
            self.tank_direction += 1
        elif symbol == arcade.key.A:
            self.tank_turning += 1
        elif symbol == arcade.key.D:
            self.tank_turning -= 1

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int):
        self.mouse_pos = x, y

    @property
    def correct(self):
        return self._correct

    @correct.setter
    def correct(self, correct: bool):
        """
        Move the tank's barrel between correct rotation and incorrect positions
        """
        self._correct = correct
        if correct:
            angle = get_angle_radians(
                self.tank.center_y, self.tank.center_x,
                self.mouse_pos[1], self.mouse_pos[0])

            self.barrel.position = (
                self.barrel.center_x + math.sin(angle) * TANK_BARREL_LENGTH_HALF,
                self.barrel.center_y + math.cos(angle) * TANK_BARREL_LENGTH_HALF,
            )

        else:
            self.barrel.position = self.tank.position


def main():
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    game = GameView()

    window.show_view(game)
    window.run()


if __name__ == '__main__':
    main()



"""
Shadertoy

Inputs:

uniform vec3      iResolution;           // viewport resolution (in pixels)
uniform float     iTime;                 // shader playback time (in seconds)
uniform float     iTimeDelta;            // render time (in seconds)
uniform int       iFrame;                // shader playback frame
uniform float     iChannelTime[4];       // channel playback time (in seconds)
uniform vec3      iChannelResolution[4]; // channel resolution (in pixels)
uniform vec4      iMouse;                // mouse pixel coords. xy: current (if MLB down), zw: click
uniform samplerXX iChannel0..3;          // input channel. XX = 2D/Cube
uniform vec4      iDate;                 // (year, month, day, time in seconds)
uniform float     iSampleRate;           // sound sample rate (i.e., 44100)
"""

import string
from datetime import datetime
from pathlib import Path

import arcade
from arcade import get_window
from arcade.context import ArcadeContext
from arcade.gl import Texture2D, geometry
from arcade.gl.framebuffer import Framebuffer
from arcade.gl.program import Program


class ShadertoyBase:
    """
    Base class for shadertoy types.
    It can represent the main image or buffers.

    Supported uniforms are::

        uniform float     iTime;                 // shader playback time (in seconds)
        uniform float     iChannelTime[4];       // channel playback time (in seconds)
        uniform vec4      iMouse;                // mouse pixel coords. xy: current (if MLB down)
                                                   zw: click
        uniform vec3      iResolution;           // viewport resolution (in pixels)
        uniform vec3      iChannelResolution[4]; // channel resolution (in pixels)
        uniform int       iFrame;                // shader playback frame
        uniform float     iTimeDelta;            // render time (in seconds)
        uniform vec4      iDate;                 // (year, month, day, time in seconds)
        // Channel textures
        uniform sampler2D iChannel0;
        uniform sampler2D iChannel1;
        uniform sampler2D iChannel2;
        uniform sampler2D iChannel3;

    Args:
        size:
            screen/area size
        source:
            The mainImage shader source
    """

    def __init__(self, size: tuple[int, int], source: str):
        self._ctx = get_window().ctx
        self._size = size
        self._source = source
        # Uniforms
        self._mouse_pos = 0.0, 0.0
        self._mouse_buttons = 0.0, 0.0
        self._time: float = 0.0
        self._time_delta: float = 0.0
        self._frame: int = 0
        self._frame_rate: float = 0.0
        self._channel_time = [0.0, 0.0, 0.0, 0.0]
        self._channel_resolution = [0] * 3 * 4
        # Shader inputs
        self._channel_0: Texture2D | None = None
        self._channel_1: Texture2D | None = None
        self._channel_2: Texture2D | None = None
        self._channel_3: Texture2D | None = None

        self._set_source(source)
        self._quad = geometry.quad_2d_fs()

    @property
    def size(self) -> tuple[int, int]:
        """
        Get or set the size in pixels.

        Mapped to uniform ``iResolution.xy``.
        """
        return self._size

    @size.setter
    def size(self, value):
        self._size = value

    @property
    def time(self) -> float:
        """
        Get or set the current time.

        Mapped to uniform ``iTime``.
        """
        return self._time

    @time.setter
    def time(self, value):
        self._time = value

    @property
    def time_delta(self) -> float:
        """
        Get or set the current delta time.

        Mapped to uniform ``iTimeDelta``.
        """
        return self._time_delta

    @time_delta.setter
    def time_delta(self, value):
        self._time_delta = value

    @property
    def delta_time(self) -> float:
        """
        Get or set the current delta time.
        An alternative to ``time_delta``.

        Mapped to uniform ``iTimeDelta``.
        """
        return self._time_delta

    @delta_time.setter
    def delta_time(self, value):
        self._time_delta = value

    @property
    def frame(self) -> int:
        """
        Get or set the current frame.

        Mapped to uniform ``iFrame``.
        """
        return self._frame

    @frame.setter
    def frame(self, value):
        self._frame = value

    @property
    def frame_rate(self) -> float:
        """
        Get or set the frame rate.

        Mapped to uniform ``iFrameRate``.
        """
        return self._frame_rate

    @frame_rate.setter
    def frame_rate(self, value: float):
        self._frame_rate = value

    @property
    def mouse_position(self) -> tuple[float, float]:
        """
        Get or set the current mouse position.

        Mapped to uniform ``iMouse.xy``.
        """
        return self._mouse_pos

    @mouse_position.setter
    def mouse_position(self, value):
        self._mouse_pos = value

    @property
    def mouse_buttons(self) -> tuple[float, float]:
        """
        Get or set the mouse button states.
        Depending on the use case these can contain
        a non-zero value when buttons are pushed and/or
        the actual click position.

        Mapped to uniform ``iMouse.zw``.
        """
        return self._mouse_buttons

    @mouse_buttons.setter
    def mouse_buttons(self, value: tuple[float, float]):
        self._mouse_buttons = value

    @property
    def channel_time(self) -> list[float]:
        return self._channel_time

    @property
    def channel_0(self) -> Texture2D | None:
        """Get or set channel 0"""
        return self._channel_0

    @channel_0.setter
    def channel_0(self, value: Texture2D):
        if not isinstance(value, Texture2D):
            raise ValueError("A channel only accepts an arcade.gl.Texture")
        self._channel_resolution[0:3] = value.width, value.height, 1
        self._channel_0 = value

    @property
    def channel_1(self) -> Texture2D | None:
        """Get or set channel 1"""
        return self._channel_1

    @channel_1.setter
    def channel_1(self, value: Texture2D):
        if not isinstance(value, Texture2D):
            raise ValueError("A channel only accepts an arcade.gl.Texture")
        self._channel_resolution[3:6] = value.width, value.height, 1
        self._channel_1 = value

    @property
    def channel_2(self) -> Texture2D | None:
        """Get or set channel 2"""
        return self._channel_2

    @channel_2.setter
    def channel_2(self, value: Texture2D):
        if not isinstance(value, Texture2D):
            raise ValueError("A channel only accepts an arcade.gl.Texture")
        self._channel_resolution[6:9] = value.width, value.height, 1
        self._channel_2 = value

    @property
    def channel_3(self) -> Texture2D | None:
        """Get or set channel 3"""
        return self._channel_3

    @channel_3.setter
    def channel_3(self, value: Texture2D):
        if not isinstance(value, Texture2D):
            raise ValueError("A channel only accepts an arcade.gl.Texture")
        self._channel_resolution[9:12] = value.width, value.height, 1
        self._channel_3 = value

    @property
    def program(self) -> Program:
        """The shader program"""
        return self._program

    @property
    def ctx(self) -> ArcadeContext:
        """The context"""
        return self._ctx

    def resize(self, size: tuple[int, int]) -> None:
        """Resize of this shadertoy or buffer"""
        raise NotImplementedError

    def render(
        self,
        *,
        time: float | None = None,
        time_delta: float | None = None,
        mouse_position: tuple[float, float] | None = None,
        size: tuple[int, int] | None = None,
        frame: int | None = None,
        frame_rate: float | None = None,
    ):
        """
        Render the shadertoy project to the screen.

        Args:
            time:
                Override the time
            time_delta:
                Override the time delta
            mouse_position:
                Override mouse position
            size:
                Override the size
            frame:
                Override frame
            frame_rate:
                Override frame_rate
        """
        self._time = time if time is not None else self._time
        self._time_delta = time_delta if time_delta is not None else self._time_delta
        self._mouse_pos = mouse_position if mouse_position is not None else self._mouse_pos
        self._size = size if size is not None else self._size
        self._frame = frame if frame is not None else self._frame
        self._frame_rate = frame_rate if frame_rate is not None else self._frame_rate
        self._render()

    def _render(self):
        raise NotImplementedError

    def reload(self, source: str):
        """
        Update the shader source code.

        Args:
            source: New mainImage shader source
        """
        self._set_source(source)

    def _bind_channels(self):
        """Bind channel textures if set"""
        if self._channel_0:
            self._channel_0.use(0)
        if self._channel_1:
            self._channel_1.use(1)
        if self._channel_2:
            self._channel_2.use(2)
        if self._channel_3:
            self._channel_3.use(3)

    def _set_uniforms(self):
        """Attempt to set all supported uniforms"""
        self._program.set_uniform_safe("iTime", self._time)
        self._program.set_uniform_array_safe("iChannelTime", self._channel_time)
        self._program.set_uniform_safe("iTimeDelta", self._time_delta)
        self._program.set_uniform_safe("iMouse", (*self._mouse_pos, *self._mouse_buttons))
        self._program.set_uniform_safe("iResolution", (*self._size, 1.0))
        self._program.set_uniform_array_safe("iChannelResolution", self._channel_resolution)
        self._program.set_uniform_safe("iFrame", self._frame)
        self._program.set_uniform_safe("iFrameRate", self._frame_rate)
        self._program.set_uniform_safe("iDate", self._get_date())

    def _get_date(self) -> tuple[float, float, float, float]:
        """Create year, month, day, seconds data for iDate"""
        now = datetime.now()
        seconds = now.hour * 60 * 60 + now.minute * 60 + now.second + now.microsecond / 1_000_000
        return now.year, now.month, now.day, seconds

    def _set_source(self, source: str):
        """
        Load shader templates, injects main function,
        create program and configures the program.
        """
        vs_path = arcade.resources.resolve(":system:shaders/shadertoy/base_vs.glsl")
        fs_path = arcade.resources.resolve(":system:shaders/shadertoy/base_fs.glsl")
        with open(vs_path) as fd:
            vs_source = fd.read()
        with open(fs_path) as fd:
            fs_source = fd.read()

        template = string.Template(fs_source)

        self._program = self.ctx.program(
            vertex_shader=vs_source,
            fragment_shader=template.substitute({"mainfunc": source}),
        )
        # Configure texture channels
        for channel in [0, 1, 2, 3]:
            try:
                self._program[f"iChannel{channel}"] = channel
            except KeyError:
                pass
        self._source = source


class ShadertoyBuffer(ShadertoyBase):
    """
    An offscreen framebuffer we can render to with the supplied
    shader or render any other content into.

    Args:
        size:
            Size of framebuffer / texture
        source:
            mainImage shader source
        repeat:
            Repeat/wrap mode for the underlying texture
    """

    def __init__(self, size: tuple[int, int], source: str, repeat: bool = False):
        super().__init__(size, source)
        self._texture = self.ctx.texture(self._size, components=4)
        self._fbo = self.ctx.framebuffer(color_attachments=[self._texture])
        self._repeat = repeat
        self._set_repeat()

    @property
    def texture(self) -> Texture2D:
        """
        The OpenGL texture for this buffer.
        This can be assigned to channels.
        """
        return self._texture

    @property
    def fbo(self) -> Framebuffer:
        """The framebuffer for this buffer"""
        return self._fbo

    @property
    def repeat(self) -> bool:
        """
        Get or set texture repeat.
        """
        return self._repeat

    @repeat.setter
    def repeat(self, value: bool):
        self._repeat = value
        self._set_repeat()

    def clear(self):
        """Clear the buffer contents"""
        self._fbo.clear()

    def _set_repeat(self):
        if self._repeat:
            self._texture.wrap_x = self.ctx.REPEAT
            self._texture.wrap_y = self.ctx.REPEAT
        else:
            self._texture.wrap_x = self.ctx.CLAMP_TO_EDGE
            self._texture.wrap_y = self.ctx.CLAMP_TO_EDGE

    def _render(self):
        self._bind_channels()
        self._set_uniforms()
        with self._fbo.activate():
            self._quad.render(self._program)

    def resize(self, size: tuple[int, int]):
        """
        Change the internal buffer size.

        Args:
            size: New size
        """
        if self._size == size:
            return
        self._size = size
        # Resize the internal texture and fbo + clear
        self._texture.resize(self._size)
        self._fbo.resize()
        self._fbo.clear()


class Shadertoy(ShadertoyBase):
    """A ShaderToy interface for arcade.

    Simply implement the ``mainImage`` glsl method::

        void mainImage( out vec4 fragColor, in vec2 fragCoord ) {
            fragColor = vec4(fragCoord, 0.0, 1.0);
        }

    Args:
        size: pixel size if the output
        main_source: The main glsl source with mainImage function
    """

    def __init__(self, size: tuple[int, int], main_source: str):
        super().__init__(size, main_source)

        self._buffer_a: ShadertoyBuffer | None = None
        self._buffer_b: ShadertoyBuffer | None = None
        self._buffer_c: ShadertoyBuffer | None = None
        self._buffer_d: ShadertoyBuffer | None = None

    @property
    def buffer_a(self) -> ShadertoyBuffer | None:
        """Get or set buffer a"""
        return self._buffer_a

    @buffer_a.setter
    def buffer_a(self, value):
        self._buffer_a = value

    @property
    def buffer_b(self) -> ShadertoyBuffer | None:
        """Get or set buffer b"""
        return self._buffer_b

    @buffer_b.setter
    def buffer_b(self, value):
        self._buffer_b = value

    @property
    def buffer_c(self) -> ShadertoyBuffer | None:
        """Get or set buffer c"""
        return self._buffer_c

    @buffer_c.setter
    def buffer_c(self, value):
        self._buffer_c = value

    @property
    def buffer_d(self) -> ShadertoyBuffer | None:
        """Get or set buffer d"""
        return self._buffer_d

    @buffer_d.setter
    def buffer_d(self, value):
        self._buffer_d = value

    @classmethod
    def create_from_file(cls, size: tuple[int, int], path: str | Path) -> "Shadertoy":
        """
        Create a Shadertoy from a mainImage shader file.

        Args:
            size: Size of shadertoy in pixels
            path: Path to mainImage shader file
        """
        path = arcade.resources.resolve(path)
        with open(path) as fd:
            source = fd.read()
        return cls(size, source)

    def create_buffer(self, source: str, repeat: bool = False) -> ShadertoyBuffer:
        """
        Shortcut for creating a buffer from mainImage shader file.

        Args:
            source: Path to shader file
            repeat: Buffer/texture repeat at borders
        """
        return ShadertoyBuffer(self._size, source, repeat=repeat)

    def create_buffer_from_file(self, path: str | Path) -> ShadertoyBuffer:
        """
        Shortcut for creating a ShadertoyBuffer from shaders source.
        The size of the framebuffer will be the same as the Shadertoy.

        Args:
            path: Path to shader source
        """
        path = arcade.resources.resolve(path)
        with open(path) as fd:
            source = fd.read()
        return ShadertoyBuffer(self._size, source)

    def resize(self, size: tuple[int, int]):
        """
        Resize the internal buffers

        Args:
            size: The new size in pixels
        """
        self._size = size

        if self._buffer_a:
            self._buffer_a.resize(size)
        if self._buffer_b:
            self._buffer_b.resize(size)
        if self._buffer_c:
            self._buffer_c.resize(size)
        if self._buffer_d:
            self._buffer_d.resize(size)

    def _render(self):
        # Render buffers first
        # with self.ctx.enabled_only():
        buffers = [self._buffer_a, self._buffer_b, self._buffer_c, self._buffer_d]
        for buffer in buffers:
            if buffer is not None:
                buffer.render(
                    time=self._time,
                    time_delta=self._time_delta,
                    mouse_position=self._mouse_pos,
                    size=self._size,
                    frame=self._frame,
                )

        # Run the main program
        self._bind_channels()
        self._set_uniforms()
        self._quad.render(self._program)



"""
Loads the Arcade version into a Python-readable ``VERSION`` string.

Everyday Arcade users may prefer accessing the ``VERSION`` string
from Arcade's top-level alias:

.. code-block:: python

   import sys
   import arcade

   if arcade.version < "3.0.0":
       # Using file=sys.stderr prints to the error stream (usually prints red)
       print("This game requires Arcade 3.0.0+ to run!", file=sys.stderr)

The ``VERSION`` constant in this module will be loaded from a file in the
same directory. It will contain the following:

#. major
#. minor
#. point
#. (Optional) one and _only_ one of:

   * .dev{DEV_PREVIEW_NUMBER}
   * rc{RC_NUMBER}

"""

import re
import sys
from pathlib import Path
from typing import Final

_HERE = Path(__file__).parent

# Grab version numbers + optional dev point preview
# Assumes:
# {MAJOR}.{MINOR}.{POINT} format
# optional: one and ONLY one of:
# 1. dev{DEV_PREVIEW}
# 2. rc{RC_NUMBER}
# Q: Why did you use regex?!
# A: If the dev_preview field is invalid, the whole match fails instantly
_VERSION_REGEX: Final[re.Pattern] = re.compile(
    r"""
    # First three version number fields
      (?P<major>[0-9]+)
    \.(?P<minor>0|[1-9][0-9]*)
    \.(?P<point>0|[1-9][0-9]*)
    # Optional and mutually exclusive: dev preview or rc number
    (?:
        \.(?P<dev_preview>dev(?:0|[1-9][0-9]*))
        |  # XOR: can't be both a preview and an rc
        (?P<rc_number>rc(?:0|[1-9][0-9]*))
    )?
    """,
    re.X,
)


def _parse_python_friendly_version(
    raw_version: str, pattern: re.Pattern[str] = _VERSION_REGEX
) -> str:
    """Read a GitHub CI version string to a Python-friendly one.

    For example, ``3.1.0-dev.1`` would become ``3.1.0.dev1``.

    Args:
        raw_version:
            A raw GitHub CI version string, as read from a file.
    Returns:
        A Python-friendly version string.
    """
    # Quick preflight check: we don't support tuple format here!
    problem = None
    if not isinstance(raw_version, str):
        problem = TypeError
    elif (match := pattern.fullmatch(raw_version)) is None:
        problem = ValueError
    if problem:
        raise problem(
            f"{raw_version=!r} not a str of the format MAJOR.MINOR"
            f"POINT with at most one of dev{{DEV_PREVIEW}} or"
            f"rc{{RC_NUMBER}},"
        )

    # Build final output, including a dev preview version if present
    group_dict: dict[str, str | None] = match.groupdict()  # type: ignore
    parts: list[str] = [group_dict[k] for k in ("major", "minor", "point")]  # type: ignore
    dev_preview, rc_number = (group_dict[k] for k in ("dev_preview", "rc_number"))

    if dev_preview and rc_number:
        raise ValueError(f"Can't have both {dev_preview=!r} and {rc_number=!r}")
    elif dev_preview:
        parts.append(dev_preview)

    joined = ".".join(parts)
    if rc_number:
        joined += rc_number

    return joined


def _parse_py_version_from_file(
    version_path: str | Path = _HERE / "VERSION", write_errors_to=sys.stderr
) -> str:
    """Read & validate the VERSION file as from a limited subset.

    On failure, it will:

    #. Print an error to stderr
    #. Return "0.0.0"
    #. (Indirectly) cause any PyPI uploads to fail

    Args:
        version_path:
            The VERSION file's path, defaulting to the same directory as
            this file.
        write_errors_to:
            Makes CI simpler by allowing a stream mock to be passed easily.
    Returns:
        Either a converted version or "0.0.0" on failure.
    """
    data = "0.0.0"
    try:
        raw = Path(version_path).resolve().read_text().strip()
        data = _parse_python_friendly_version(raw)
    except Exception as e:
        print(
            f"ERROR: Unable to load version number via '{str(version_path)}': {e}",
            file=write_errors_to,
        )

    return data


VERSION: Final[str] = _parse_py_version_from_file()
"""A Python-friendly version string.

This value is converted from the GitHub-style ``VERSION`` file at the
top-level of the arcade module.
"""



"""
Arcade's version of the OpenGL Context.
Contains pre-loaded programs
"""

from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any

import pyglet
from PIL import Image
from pyglet import gl
from pyglet.graphics.shader import UniformBufferObject
from pyglet.math import Mat4

import arcade
from arcade.camera import Projector
from arcade.camera.default import DefaultProjector
from arcade.gl import BufferDescription, Context
from arcade.gl.compute_shader import ComputeShader
from arcade.gl.framebuffer import Framebuffer
from arcade.gl.program import Program
from arcade.gl.texture import Texture2D
from arcade.gl.types import PyGLenum
from arcade.gl.vertex_array import Geometry
from arcade.texture_atlas import DefaultTextureAtlas, TextureAtlasBase

__all__ = ["ArcadeContext"]


class ArcadeContext(Context):
    """
    An OpenGL context implementation for Arcade with added custom features.
    This context is normally accessed through :py:attr:`arcade.Window.ctx`.

    Args:
        window:
            The pyglet window
        gc_mode:
            The garbage collection mode for OpenGL objects. ``auto`` is just
            what we would expect in python while ``context_gc`` (default)
            requires you to call ``Context.gc()``. The latter can be useful
            when using multiple threads when it's not clear what thread will
            gc the object.
        gl_api:
            The OpenGL API to use. By default it's set to ``gl`` which is
            the standard OpenGL API. If you want to use OpenGL ES you can
            set it to ``gles``.
    """

    atlas_size: tuple[int, int] = 512, 512

    def __init__(
        self,
        window: pyglet.window.Window,  # type: ignore
        gc_mode: str = "context_gc",
        gl_api: str = "gl",
    ) -> None:
        super().__init__(window, gc_mode=gc_mode, gl_api=gl_api)

        # Set up a default orthogonal projection for sprites and shapes
        self._window_block: UniformBufferObject = window.ubo
        self.bind_window_block()

        self.blend_func = self.BLEND_DEFAULT

        self._default_camera: DefaultProjector = DefaultProjector(context=self)
        self.current_camera: Projector = self._default_camera

        self.viewport = (0, 0, window.width, window.height)

        # --- Pre-load system shaders here ---
        # FIXME: These pre-created resources needs to be packaged nicely
        #        Just having them globally in the context is probably not a good idea
        self.line_vertex_shader: Program = self.load_program(
            vertex_shader=":system:shaders/shapes/line/line_vertex_shader_vs.glsl",
            fragment_shader=":system:shaders/shapes/line/line_vertex_shader_fs.glsl",
        )
        self.line_generic_with_colors_program: Program = self.load_program(
            vertex_shader=":system:shaders/shapes/line/line_generic_with_colors_vs.glsl",
            fragment_shader=":system:shaders/shapes/line/line_generic_with_colors_fs.glsl",
        )
        self.shape_element_list_program: Program = self.load_program(
            vertex_shader=":system:shaders/shape_element_list_vs.glsl",
            fragment_shader=":system:shaders/shape_element_list_fs.glsl",
        )
        self.sprite_list_program_no_cull: Program = self.load_program(
            vertex_shader=":system:shaders/sprites/sprite_list_geometry_vs.glsl",
            geometry_shader=":system:shaders/sprites/sprite_list_geometry_no_cull_geo.glsl",
            fragment_shader=":system:shaders/sprites/sprite_list_geometry_fs.glsl",
        )
        self.sprite_list_program_no_cull["sprite_texture"] = 0
        self.sprite_list_program_no_cull["uv_texture"] = 1

        self.sprite_list_program_cull: Program = self.load_program(
            vertex_shader=":system:shaders/sprites/sprite_list_geometry_vs.glsl",
            geometry_shader=":system:shaders/sprites/sprite_list_geometry_cull_geo.glsl",
            fragment_shader=":system:shaders/sprites/sprite_list_geometry_fs.glsl",
        )
        self.sprite_list_program_cull["sprite_texture"] = 0
        self.sprite_list_program_cull["uv_texture"] = 1

        self.sprite_program_single = self.load_program(
            vertex_shader=":system:shaders/sprites/sprite_single_vs.glsl",
            geometry_shader=":system:shaders/sprites/sprite_list_geometry_no_cull_geo.glsl",
            fragment_shader=":system:shaders/sprites/sprite_list_geometry_fs.glsl",
        )
        self.sprite_program_single["sprite_texture"] = 0
        self.sprite_program_single["uv_texture"] = 1
        self.sprite_program_single["spritelist_color"] = 1.0, 1.0, 1.0, 1.0

        # Shapes
        self.shape_line_program: Program = self.load_program(
            vertex_shader=":system:shaders/shapes/line/unbuffered_vs.glsl",
            fragment_shader=":system:shaders/shapes/line/unbuffered_fs.glsl",
            geometry_shader=":system:shaders/shapes/line/unbuffered_geo.glsl",
        )
        self.shape_ellipse_filled_unbuffered_program: Program = self.load_program(
            vertex_shader=":system:shaders/shapes/ellipse/filled_unbuffered_vs.glsl",
            fragment_shader=":system:shaders/shapes/ellipse/filled_unbuffered_fs.glsl",
            geometry_shader=":system:shaders/shapes/ellipse/filled_unbuffered_geo.glsl",
        )
        self.shape_ellipse_outline_unbuffered_program: Program = self.load_program(
            vertex_shader=":system:shaders/shapes/ellipse/outline_unbuffered_vs.glsl",
            fragment_shader=":system:shaders/shapes/ellipse/outline_unbuffered_fs.glsl",
            geometry_shader=":system:shaders/shapes/ellipse/outline_unbuffered_geo.glsl",
        )
        self.shape_rectangle_filled_unbuffered_program = self.load_program(
            vertex_shader=":system:shaders/shapes/rectangle/filled_unbuffered_vs.glsl",
            fragment_shader=":system:shaders/shapes/rectangle/filled_unbuffered_fs.glsl",
            geometry_shader=":system:shaders/shapes/rectangle/filled_unbuffered_geo.glsl",
        )
        # Atlas shaders
        self.atlas_resize_program: Program = self.load_program(
            vertex_shader=":system:shaders/atlas/resize_vs.glsl",
            geometry_shader=":system:shaders/atlas/resize_gs.glsl",
            fragment_shader=":system:shaders/atlas/resize_fs.glsl",
        )
        self.atlas_resize_program["atlas_old"] = 0  # Configure texture channels
        self.atlas_resize_program["atlas_new"] = 1
        self.atlas_resize_program["texcoords_old"] = 2
        self.atlas_resize_program["texcoords_new"] = 3

        # SpriteList collision resources
        self.collision_detection_program = self.load_program(
            vertex_shader=":system:shaders/collision/col_trans_vs.glsl",
            geometry_shader=":system:shaders/collision/col_trans_gs.glsl",
        )
        self.collision_buffer = self.buffer(reserve=1024 * 4)
        self.collision_query = self.query(samples=False, time=False, primitives=True)

        # General Utility

        # renders a quad (without projection) with a single 4-component texture.
        self.utility_textured_quad_program: Program = self.load_program(
            vertex_shader=":system:shaders/util/textured_quad_vs.glsl",
            fragment_shader=":system:shaders/util/textured_quad_fs.glsl",
        )

        # --- Pre-created geometry and buffers for unbuffered draw calls ----
        # FIXME: These pre-created resources needs to be packaged nicely
        #        Just having them globally in the context is probably not a good idea
        self.generic_draw_line_strip_color = self.buffer(reserve=4 * 1000)
        self.generic_draw_line_strip_vbo = self.buffer(reserve=8 * 1000)
        self.generic_draw_line_strip_geometry = self.geometry(
            [
                BufferDescription(self.generic_draw_line_strip_vbo, "2f", ["in_vert"]),
                BufferDescription(
                    self.generic_draw_line_strip_color,
                    "4f1",
                    ["in_color"],
                ),
            ]
        )
        # Shape line(s)
        # Reserve space for 1000 lines (2f pos, 4f color)
        # TODO: Different version for buffered and unbuffered
        # TODO: Make round-robin buffers
        self.shape_line_buffer_pos = self.buffer(reserve=8 * 10)
        # self.shape_line_buffer_color = self.buffer(reserve=4 * 10)
        self.shape_line_geometry = self.geometry(
            [
                BufferDescription(self.shape_line_buffer_pos, "2f", ["in_vert"]),
                # BufferDescription(self.shape_line_buffer_color, '4f1', ['in_color'])
            ]
        )
        # ellipse/circle filled
        self.shape_ellipse_unbuffered_buffer = self.buffer(reserve=8)
        self.shape_ellipse_unbuffered_geometry: Geometry = self.geometry(
            [BufferDescription(self.shape_ellipse_unbuffered_buffer, "2f", ["in_vert"])]
        )
        # ellipse/circle outline
        self.shape_ellipse_outline_unbuffered_buffer = self.buffer(reserve=8)
        self.shape_ellipse_outline_unbuffered_geometry: Geometry = self.geometry(
            [BufferDescription(self.shape_ellipse_outline_unbuffered_buffer, "2f", ["in_vert"])]
        )
        # rectangle filled
        self.shape_rectangle_filled_unbuffered_buffer = self.buffer(reserve=8)
        self.shape_rectangle_filled_unbuffered_geometry: Geometry = self.geometry(
            [BufferDescription(self.shape_rectangle_filled_unbuffered_buffer, "2f", ["in_vert"])]
        )
        self.geometry_empty: Geometry = self.geometry()

        self._atlas: TextureAtlasBase | None = None
        # Global labels we modify in `arcade.draw_text`.
        # These multiple labels with different configurations are stored
        self.label_cache: dict[str, arcade.Text] = {}

        # self.active_program = None
        self.point_size = 1.0

    def reset(self) -> None:
        """
        Reset context flags and other states.
        This is mostly used in unit testing.
        """
        self.screen.use(force=True)
        self.bind_window_block()
        # self.active_program = None
        self.viewport = 0, 0, self.window.width, self.window.height
        self.view_matrix = Mat4()
        self.projection_matrix = Mat4.orthogonal_projection(
            0, self.window.width, 0, self.window.height, -100, 100
        )
        self.enable_only(self.BLEND)
        self.blend_func = self.BLEND_DEFAULT
        self.point_size = 1.0

    def bind_window_block(self) -> None:
        """
        Binds the global projection and view uniform buffer object.

        This should always be bound to index 0 so all shaders
        have access to them.
        """
        gl.glBindBufferRange(
            gl.GL_UNIFORM_BUFFER,
            0,
            self._window_block.buffer.id,
            0,
            128,  # 32 x 32bit floats (two mat4)
        )

    @property
    def default_atlas(self) -> TextureAtlasBase:
        """
        The default texture atlas.

        This is created when Arcade is initialized.
        All sprite lists will use use this atlas unless a different atlas
        is passed in the :py:class:`arcade.SpriteList` constructor.
        """
        if not self._atlas:
            # Create the default texture atlas
            # 8192 is a safe maximum size for textures in OpenGL 3.3
            # We might want to query the max limit, but this makes it consistent
            # across all OpenGL implementations.
            self._atlas = DefaultTextureAtlas(
                self.atlas_size,
                border=2,
                auto_resize=True,
                ctx=self,
            )

        return self._atlas

    @property
    def viewport(self) -> tuple[int, int, int, int]:
        """
        Get or set the viewport for the currently active framebuffer.

        The viewport simply describes what pixels of the screen
        OpenGL should render to. Format is ``(x, y, width, height)``.
        Normally it would be the size of the window's framebuffer::

            # 4:3 screen
            ctx.viewport = 0, 0, 800, 600
            # 1080p
            ctx.viewport = 0, 0, 1920, 1080
            # Using the current framebuffer size
            ctx.viewport = 0, 0, *ctx.screen.size
        """
        return self.active_framebuffer.viewport

    @viewport.setter
    def viewport(self, value: tuple[int, int, int, int]):
        self.active_framebuffer.viewport = value
        if self._default_camera == self.current_camera:
            self._default_camera.use()

    @property
    def projection_matrix(self) -> Mat4:
        """
        Get or set the current projection matrix.

        This 4x4 float32 matrix is usually calculated by a cameras but
        can be modified directly if you know what you are doing.

        This property simply gets and sets pyglet's projection matrix.
        """
        return self.window.projection

    @projection_matrix.setter
    def projection_matrix(self, value: Mat4):
        if not isinstance(value, Mat4):
            raise ValueError("projection_matrix must be a Mat4 object")

        self.window.projection = value

    @property
    def view_matrix(self) -> Mat4:
        """
        Get or set the current view matrix.

        This 4x4 float32 matrix is usually calculated by a cameras but
        can be modified directly if you know what you are doing.

        This property simply gets and sets pyglet's view matrix.
        """
        return self.window.view

    @view_matrix.setter
    def view_matrix(self, value: Mat4):
        if not isinstance(value, Mat4):
            raise ValueError("view_matrix must be a Mat4 object")

        self.window.view = value

    def load_program(
        self,
        *,
        vertex_shader: str | Path,
        fragment_shader: str | Path | None = None,
        geometry_shader: str | Path | None = None,
        tess_control_shader: str | Path | None = None,
        tess_evaluation_shader: str | Path | None = None,
        common: Iterable[str | Path] = (),
        defines: dict[str, Any] | None = None,
        varyings: Sequence[str] | None = None,
        varyings_capture_mode: str = "interleaved",
    ) -> Program:
        """
        Create a new program given file names that contain the vertex shader and
        fragment shader. Note that the fragment and geometry shaders are optional
        when transform shaders are loaded.

        This method also supports resource handles.

        Example::

            # The most common use case is having a vertex and fragment shader
            program = window.ctx.load_program(
                vertex_shader="vert.glsl",
                fragment_shader="frag.glsl",
            )

        Args:
            vertex_shader:
                Path to the vertex shader.
            fragment_shader:
                Path to the fragment shader.
            geometry_shader:
                Path to the geometry shader.
            tess_control_shader:
                Tessellation Control Shader.
            tess_evaluation_shader:
                Tessellation Evaluation Shader.
            common:
                Common files to be included in all shaders.
            defines:
                Substitute `#define` values in the source.
            varyings:
                The name of the out attributes in a transform shader.
                This is normally not necessary since we auto detect them,
                but some more complex out structures we can't detect.
            varyings_capture_mode:
                The capture mode for transforms.

                Based on these settings, the `transform()` method will accept a single
                buffer or a list of buffers.

                - ``"interleaved"`` means all out attributes will be written to a single buffer.
                - ``"separate"`` means each out attribute will be written to separate buffers.
        """
        from arcade.resources import resolve

        vertex_shader_src = resolve(vertex_shader).read_text()
        vertex_shader_src = self.shader_inc(vertex_shader_src)

        fragment_shader_src = None
        geometry_shader_src = None
        tess_control_src = None
        tess_evaluation_src = None

        common_src = [resolve(c).read_text() for c in common]

        if fragment_shader:
            fragment_shader_src = resolve(fragment_shader).read_text()
            fragment_shader_src = self.shader_inc(fragment_shader_src)

        if geometry_shader:
            geometry_shader_src = resolve(geometry_shader).read_text()
            geometry_shader_src = self.shader_inc(geometry_shader_src)

        if tess_control_shader and tess_evaluation_shader:
            tess_control_src = resolve(tess_control_shader).read_text()
            tess_evaluation_src = resolve(tess_evaluation_shader).read_text()
            tess_control_src = self.shader_inc(tess_control_src)
            tess_evaluation_src = self.shader_inc(tess_evaluation_src)

        return self.program(
            vertex_shader=vertex_shader_src,
            fragment_shader=fragment_shader_src,
            geometry_shader=geometry_shader_src,
            tess_control_shader=tess_control_src,
            tess_evaluation_shader=tess_evaluation_src,
            common=common_src,
            defines=defines,
            varyings=varyings,
            varyings_capture_mode=varyings_capture_mode,
        )

    def load_compute_shader(
        self, path: str | Path, common: Iterable[str | Path] = ()
    ) -> ComputeShader:
        """
        Loads a compute shader from file. This methods supports
        resource handles.

        Example::

            ctx.load_compute_shader(":shader:compute/do_work.glsl")

        Args:
            path:
                Path to texture
            common:
                Common sources injected into compute shader
        """
        from arcade.resources import resolve

        path = resolve(path)
        common_src = [resolve(c).read_text() for c in common]
        return self.compute_shader(
            source=self.shader_inc(path.read_text()),
            common=common_src,
        )

    def load_texture(
        self,
        path: str | Path,
        *,
        flip: bool = True,
        wrap_x: PyGLenum | None = None,
        wrap_y: PyGLenum | None = None,
        filter: tuple[PyGLenum, PyGLenum] | None = None,
        build_mipmaps: bool = False,
        internal_format: int | None = None,
        immutable: bool = False,
        compressed: bool = False,
    ) -> Texture2D:
        """
        Loads and creates an OpenGL 2D texture.
        Currently, all textures are converted to RGBA for simplicity.

        Examples::

            # Load a texture in current working directory
            texture = window.ctx.load_texture("background.png")

            # Load a texture using Arcade resource handle
            texture = window.ctx.load_texture(":textures:background.png")

            # Load and compress a texture
            texture = window.ctx.load_texture(
                ":textures:background.png",
                internal_format=gl.GL_COMPRESSED_RGBA_S3TC_DXT5_EXT,
                compressed=True,
            )

        Args:
            path:
                Path to texture
            flip:
                Flips the image upside down. Default is ``True``.
            wrap_x:
                The wrap mode for the x-axis. Default is ``None``.
            wrap_y:
                The wrap mode for the y-axis. Default is ``None``.
            filter:
                The min and mag filter. Default is ``None``.
            build_mipmaps:
                Build mipmaps for the texture. Default is ``False``.
            internal_format:
                The internal format of the texture. This can be used to override
                the default internal format when using sRGBA or compressed textures.
            immutable:
                Make the storage (not the contents) immutable. This can sometimes be
                required when using textures with compute shaders.
            compressed:
                If the internal format is a compressed format meaning your
                texture will be compressed by the GPU.
        """
        from arcade.resources import resolve

        path = resolve(path)

        image: Image.Image = Image.open(str(path))  # type: ignore

        if flip:
            image = image.transpose(Image.Transpose.FLIP_TOP_BOTTOM)

        texture = self.texture(
            image.size,
            components=4,
            data=image.convert("RGBA").tobytes(),
            wrap_x=wrap_x,
            wrap_y=wrap_y,
            filter=filter,
            internal_format=internal_format,
            immutable=immutable,
            compressed=compressed,
        )
        image.close()

        if build_mipmaps:
            texture.build_mipmaps()

        return texture

    def shader_inc(self, source: str) -> str:
        """
        Parse a shader source looking for ``#include`` directives and
        replace them with the contents of the included file.

        The ``#include`` directive must be on its own line and the file
        and the path should use a resource handle.

        Example::

            #include :my_resource_handle:lib/common.glsl

        Args:
            source: The shader source code
        """
        from arcade.resources import resolve

        lines = source.splitlines()
        for i, line in enumerate(lines):
            line = line.strip()
            if line.startswith("#include"):
                path = resolve(line.split()[1].replace('"', ""))
                lines[i] = path.read_text()
        return "\n".join(lines)

    def get_framebuffer_image(
        self,
        fbo: Framebuffer,
        components: int = 4,
        flip: bool = True,
    ) -> Image.Image:
        """
        Shortcut method for reading data from a framebuffer and converting it to a PIL image.

        Args:
            fbo:
                Framebuffer to get image from
            components:
                Number of components to read. Default is 4 (RGBA).
                Valid values are 1, 2, 3, 4.
            flip:
                Flip the image upside down. This is useful because OpenGL has the
                origin at the bottom left corner while PIL has it at the top left.
        """
        mode = "RGBA"[:components]
        image = Image.frombuffer(
            mode,
            (fbo.width, fbo.height),
            fbo.read(components=components),
        )
        if flip:
            image = image.transpose(Image.Transpose.FLIP_TOP_BOTTOM)

        return image



"""
The Camera Types, Classes, and Methods of Arcade.
Providing a multitude of camera's for any need.
"""

from arcade.camera.data_types import (
    Projection,
    Projector,
    CameraData,
    OrthographicProjectionData,
    PerspectiveProjectionData,
)

from arcade.camera.projection_functions import (
    generate_view_matrix,
    generate_orthographic_matrix,
    generate_perspective_matrix,
    project_orthographic,
    project_perspective,
    unproject_orthographic,
    unproject_perspective,
)

from arcade.camera.orthographic import OrthographicProjector
from arcade.camera.perspective import PerspectiveProjector

from arcade.camera.camera_2d import Camera2D

import arcade.camera.grips as grips


__all__ = [
    "Projection",
    "Projector",
    "CameraData",
    "generate_view_matrix",
    "OrthographicProjectionData",
    "generate_orthographic_matrix",
    "project_orthographic",
    "unproject_orthographic",
    "OrthographicProjector",
    "PerspectiveProjectionData",
    "generate_perspective_matrix",
    "project_perspective",
    "unproject_perspective",
    "PerspectiveProjector",
    "Camera2D",
    "grips",
]



"""
A wrapper over OpenGL 3.3 core making OpenGL more reasonable to work with and easier
to learn. The API is based on `ModernGL <https://github.com/moderngl/moderngl>`_
implementing a subset of the features. We use pyglet's OpenGL bindings based on ctypes.

Creating OpenGL resources such as buffers, framebuffers, programs (shaders) and textures
should be done through methods in a context.

* Arcade users should access :py:attr:`arcade.Window.ctx` exposing an
  :py:class:`arcade.ArcadeContext`
* Pyglet users can instantiate an :py:class:`arcade.gl.Context` for the window or
  extend this class with more features if needed.

.. warning:: This module contains the low level rendering API for arcade
             and is only recommended for more advanced users
"""

from .context import Context
from .types import BufferDescription
from .compute_shader import ComputeShader
from .exceptions import ShaderException
from .enums import *  # noqa
from .buffer import Buffer
from .vertex_array import Geometry, VertexArray
from .texture import Texture2D
from .texture_array import TextureArray
from .sampler import Sampler
from .framebuffer import Framebuffer
from .program import Program
from .query import Query
from . import geometry

__all__ = [
    "Buffer",
    "BufferDescription",
    "ComputeShader",
    "Context",
    "Framebuffer",
    "Geometry",
    "Program",
    "Query",
    "ShaderException",
    "VertexArray",
    "Texture2D",
    "TextureArray",
    "Sampler",
    "geometry",
]



import PIL.Image
import PIL.ImageDraw
import PIL.ImageOps

from arcade import cache
from arcade.color import TRANSPARENT_BLACK
from arcade.hitbox import HitBoxAlgorithm
from arcade.math import lerp
from arcade.types import RGBA255

from .texture import ImageData, Texture


def make_circle_texture(
    diameter: int,
    color: RGBA255,
    name: str | None = None,
    hit_box_algorithm: HitBoxAlgorithm | None = None,
) -> Texture:
    """
    Creates a :class:`Texture` of a circle with the given diameter and color.

    Args:
        diameter:
            Diameter of the circle and dimensions of the square :class:`Texture` returned.
        color:
            Color of the circle as a :py:class:`~arcade.types.Color` instance a 3 or 4 tuple.
        name:
            A unique name for the texture. If not provided, a name will be generated.
            This is used for caching and unique identifier for texture atlases.
        hit_box_algorithm:
            The hit box algorithm to use for this texture. If not provided, the default
            hit box algorithm will be used.
    """
    name = name or cache.crate_str_from_values(
        "circle_texture", diameter, color[0], color[1], color[2]
    )
    bg_color = TRANSPARENT_BLACK  # fully transparent
    img = PIL.Image.new("RGBA", (diameter, diameter), bg_color)
    draw = PIL.ImageDraw.Draw(img)
    draw.ellipse((0, 0, diameter - 1, diameter - 1), fill=color)
    return Texture(ImageData(img, hash=name), hit_box_algorithm=hit_box_algorithm)


def make_soft_circle_texture(
    diameter: int,
    color: RGBA255,
    center_alpha: int = 255,
    outer_alpha: int = 0,
    name: str | None = None,
    hit_box_algorithm: HitBoxAlgorithm | None = None,
) -> Texture:
    """
    Creates a :class:`Texture` of a circle with the given diameter and color,
    fading out at its edges.

    Args:
        diameter:
            Diameter of the circle and dimensions of the square :class:`Texture` returned.
        color:
            Color of the circle as a 4-length tuple or :py:class:`~arcade.types.Color` instance.
        center_alpha:
            Alpha value of the circle at its center.
        outer_alpha:
            Alpha value of the circle at its edges.
        name:
            A unique name for the texture. If not provided, a name will be generated.
            This is used for caching and unique identifier for texture atlases.
        hit_box_algorithm:
            The hit box algorithm to use for this texture. If not provided, the default
            hit box algorithm will be used.
    """
    # Name must be unique for caching
    name = cache.crate_str_from_values(
        "soft_circle_texture",
        diameter,
        color[0],
        color[1],
        color[2],
        center_alpha,
        outer_alpha,
    )

    bg_color = TRANSPARENT_BLACK
    img = PIL.Image.new("RGBA", (diameter, diameter), bg_color)
    draw = PIL.ImageDraw.Draw(img)
    max_radius = int(diameter // 2)
    center = max_radius

    for radius in range(max_radius, 0, -1):
        alpha = int(lerp(center_alpha, outer_alpha, radius / max_radius))
        clr = (color[0], color[1], color[2], alpha)
        draw.ellipse(
            (
                center - radius,
                center - radius,
                center + radius - 1,
                center + radius - 1,
            ),
            fill=clr,
        )

    return Texture(ImageData(img, name), hit_box_algorithm=hit_box_algorithm)


def make_soft_square_texture(
    size: int,
    color: RGBA255,
    center_alpha: int = 255,
    outer_alpha: int = 0,
    name: str | None = None,
) -> Texture:
    """
    Creates a :class:`Texture` of a square with the given diameter and color,
    fading out at its edges.

    Args:
        size:
            Diameter of the square and dimensions of the square Texture returned.
        color:
            Color of the square.
        center_alpha:
            Alpha value of the square at its center.
        outer_alpha:
            Alpha value of the square at its edges.
        name:
            A unique name for the texture. If not provided, a name will be generated.
            This is used for caching and unique identifier for texture atlases.
    """
    # Build name used for caching
    name = name or cache.crate_str_from_values(
        "gradient-square", size, color, center_alpha, outer_alpha
    )

    # Generate the soft square image
    bg_color = TRANSPARENT_BLACK
    img = PIL.Image.new("RGBA", (size, size), bg_color)
    draw = PIL.ImageDraw.Draw(img)
    half_size = int(size // 2)

    for cur_size in range(0, half_size):
        alpha = int(lerp(outer_alpha, center_alpha, cur_size / half_size))
        clr = (color[0], color[1], color[2], alpha)
        draw.rectangle((cur_size, cur_size, size - cur_size, size - cur_size), clr, None)

    return Texture(img, name=name)



"""
Depth-sort sprites using a depth buffer in the GL context.

Press the space bar to toggle depth testing during drawing.

During each update, the depth of each sprite is updated to follow a
cosine wave. Afterward, the following is drawn:

* All sprites in depth-sorted order
* A white square centered over each sprite along the x-axis, and moving
  with the wave along the y-axis

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_depth_cosine
"""

import math

from pyglet.graphics import Batch

import arcade

# All constants are in pixels
WINDOW_WIDTH, WINDOW_HEIGHT = 1280, 720

WINDOW_TITLE = "Sprite Depth Testing Example w/ a Cosine Wave"
NUM_SPRITES = 10

SPRITE_X_START = 150
SPRITE_X_STEP = 50
SPRITE_Y = WINDOW_HEIGHT // 2

DOT_SIZE = 10


class GameView(arcade.View):

    def __init__(self):
        super().__init__()

        texture = arcade.load_texture(":resources:images/test_textures/xy_square.png")
        self.text_batch = Batch()

        self.use_depth: bool = True
        self.text_use_depth = arcade.Text(
            "SPACE: Toggle depth testing (True)",
            x=10,
            y=30,
            font_size=15,
            color=arcade.color.WHITE,
            batch=self.text_batch,
        )

        self.sprite_list = arcade.SpriteList()

        for i in range(NUM_SPRITES):
            sprite = arcade.Sprite(
                texture, center_x=SPRITE_X_START + SPRITE_X_STEP * i, center_y=SPRITE_Y
            )
            self.sprite_list.append(sprite)

    def on_draw(self):
        self.clear()

        ctx = self.window.ctx
        if self.use_depth:
            # This with block temporarily enables depth testing
            with ctx.enabled(ctx.DEPTH_TEST):
                self.sprite_list.draw()
        else:
            self.sprite_list.draw()

        # Draw wave visualization markers over each sprite
        for i, sprite in enumerate(self.sprite_list):
            arcade.draw_point(
                SPRITE_X_START + SPRITE_X_STEP * i,
                SPRITE_Y + sprite.depth,
                arcade.color.WHITE,
                DOT_SIZE,
            )

        self.text_batch.draw()

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.SPACE:
            self.use_depth = not self.use_depth
            self.text_use_depth.text = f"SPACE: Toggle depth testing ({self.use_depth})"

    def on_update(self, delta_time):
        # Using time from the window's clock simplifies the math below
        time = self.window.time
        for i, sprite in enumerate(self.sprite_list):
            sprite.depth = math.cos(time + i) * SPRITE_X_STEP


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
A simple cache for hit boxes calculated from texture.
Hit box calculations are normally done at load time
by inspecting the contents of a loaded image.

Depending on the hit box algorithm a hit box is calculated:
* None     : Simply a box around the whole texture. No caching needed.
* Simple   : Scanning the corners for the texture
* Detailed : fairly detailed hit box generated by pymunk with detail parameter
"""

from __future__ import annotations

import gzip
import json
from collections import OrderedDict
from pathlib import Path
from typing import TYPE_CHECKING

from arcade.resources import resolve
from arcade.types import Point2List

if TYPE_CHECKING:
    from arcade import Texture


class HitBoxCache:
    """
    A simple cache for hit box points for textures.

    These are calculated when loading a texture
    depending on the selected hit box algorithm.

    Points are stored as a tuple of xy points since
    it's important that they are immutable.
    """

    VERSION = 1

    def __init__(self):
        self._entries: OrderedDict[str, Point2List] = OrderedDict()

    def __len__(self) -> int:
        return len(self._entries)

    def __iter__(self):
        return iter(self._entries)

    def get(self, name_or_texture: str | Texture) -> Point2List | None:
        """
        Get the hit box points for a texture with a given hash
        and hit box algorithm.

        Example::

            # Get cached hit box for texture
            points = cache.get(texture)
            # Get a cache entry by string
            points = cache.get("hash|(0, 1, 2, 3)|simple|")

        Args:
            name_or_texture:
                The texture or cache name to get the hit box for
        """
        from arcade import Texture

        if isinstance(name_or_texture, Texture):
            return self._entries.get(name_or_texture.cache_name, None)
        elif isinstance(name_or_texture, str):
            return self._entries.get(name_or_texture, None)
        else:
            raise TypeError(f"Expected str or Texture: {name_or_texture}")

    def put(self, name_or_texture: str | Texture, points: Point2List) -> None:
        """
        Store hit box points for a texture.

        Example::

            # Cache hit box for texture
            cache.put(texture, points)
            # Cache with custom string
            cache.put("my_custom_points", points)

        Args:
            name_or_texture:
                The texture or cache name to store the hit box for
            points:
                The hit box points
        """
        from arcade import Texture

        # Points can be empty OR have at least 3 points
        if len(points) < 3 and len(points) != 0:
            raise ValueError(f"Hit box must have at least 3 points: {points}")

        if isinstance(name_or_texture, Texture):
            self._entries[name_or_texture.cache_name] = tuple(points)
        elif isinstance(name_or_texture, str):
            self._entries[name_or_texture] = tuple(points)
        else:
            raise TypeError(f"Expected str or Texture: {name_or_texture}")

    def load(self, path: str | Path) -> None:
        """
        Load a json file containing hit boxes.

        This adds the loaded hit boxes to the cache overwriting any existing
        entries and can therefore be called multiple times to populate it.

        if the file extension is ".gz" the file will be compressed.

        Args:
            path: The path to the json file to load
        """
        path = resolve(path)
        if path.suffix == ".gz":
            with gzip.open(path, mode="rb") as fd:
                data = json.loads(fd.read())
        else:
            with open(path) as fd:
                data = json.loads(fd.read())

        for key, value in data.items():
            self._entries[key] = tuple(value)

    def save(self, path: Path, indent: int = 0) -> None:
        """
        Save the hit box cache to disk.

        This can be used to pre-populate the cache to
        reduce the time it takes to load textures.

        if the file extension is ".gz" the file will be compressed.

        Args:
            path:
                The path to save the cache to
            indent:
                The indentation level for the json file
        """
        if indent == 0:
            data_str = json.dumps(self._entries)
        else:
            data_str = json.dumps(self._entries, indent=indent)

        data = data_str.encode("utf-8")

        if path.suffix == ".gz":
            data = gzip.compress(data)

        with open(path, mode="wb") as fd:
            fd.write(data)

    def flush(self) -> None:
        """Clear the cache."""
        self._entries.clear()

    def __repr__(self) -> str:
        return f"HitBoxCache(entries={len(self)})"



"""
Example using indirect rendering using simple buffer arrays.

Indirect rendering is simply describing one or multiple
rendering calls in a Buffer. This buffer can be written
to from the python side or generated by a compute shader.

The advantage of indirect rendering is that you can pack
all your objects in the same buffer(s) and the indirect
rendering call can draw all of them at once.

The Geometry.geometry_indirect method takes an additional
parameter (a Buffer) containing the info for each rendering call
to issue.

The format of this buffer depends if we are rendering with
or without index buffer.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.render_indirect
"""

from random import (
    randint,
    random,
    uniform,
)
from array import array

import arcade
from arcade.gl import BufferDescription

WINDOW_WIDTH = 800
WINDOW_HEIGHT = 600


class RenderIndirect(arcade.Window):

    def __init__(self):
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, "Render Indirect", gl_version=(4, 3))
        # We don't want vsync when using a query to measure performance
        self.set_vsync(False)
        self.frame = 0

        # Let's generate some simple geometry
        self.num_objects = 10_000
        # Draw parameters. These can be adjusted to render parts of the geometry.
        self.start = 0
        self.count = self.num_objects

        # Query object to measure the rendering calls
        self.query = self.ctx.query()

        # Generate:
        # geometry: Triangles for lots of squares
        # indices: Optional indices for an index buffer
        # array_draw_commands: List of draw commands for rendering each individual
        # square (without index buffer)
        # indexed_draw_commands: List of draw commands for rendering each individual
        # square (with index buffer)
        geometry, indices, array_draw_commands, indexed_draw_commands = self.gen_data(
            self.num_objects
        )

        # Crate OpenGL buffers of the data
        self.draw_command_array_buffer = self.ctx.buffer(data=array("I", array_draw_commands))
        self.draw_command_indexed_buffer = self.ctx.buffer(data=array("I", indexed_draw_commands))
        self.vbo = self.ctx.buffer(data=array("f", geometry))
        self.ibo = self.ctx.buffer(data=array("i", indices))

        # Our geometry without index buffer
        self.geometry_array = self.ctx.geometry(
            [BufferDescription(self.vbo, "2f 4f", ("in_pos", "in_color"))],
            mode=self.ctx.TRIANGLES,
        )
        # Our geometry with index buffer
        self.geometry_indexed = self.ctx.geometry(
            [BufferDescription(self.vbo, "2f 4f", ("in_pos", "in_color"))],
            index_buffer=self.ibo,
            mode=self.ctx.TRIANGLES,
        )

        # Simple program rendering colored geometry
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            // Piggyback on pyglet's global projection matrix
            uniform WindowBlock {
                mat4 projection;
                mat4 view;
            } window;

            in vec2 in_pos;
            in vec4 in_color;
            out vec4 color;

            void main() {
                gl_Position = window.projection * window.view * vec4(in_pos, 0.0, 1.0);
                color = in_color;
            }
            """,
            fragment_shader="""
            #version 330

            in vec4 color;
            out vec4 fragColor;

            void main() {
                fragColor = vec4(color);
            }
            """,
        )

    def gen_data(self, count):
        # Simple unit quad consisting of 2 rectangles
        quad = (
            # First triangle
            (-0.5, 0.5),  # Top left
            (-0.5, -0.5),  # Bottom left
            (0.5, 0.5),  # Top right
            # Second triangle
            (0.5, 0.5),  # Top right
            (-0.5, -0.5),  # Bottom left
            (0.5, -0.5),  # Bottom right
        )

        geometry = []
        index_buffer = []
        array_commands = []
        indexed_commands = []

        # Create count quads
        for quad_index in range(count):
            # Apply random offset, scale and colors
            offset_x = randint(0, WINDOW_WIDTH)
            offset_y = randint(0, WINDOW_HEIGHT)
            scale = randint(5, 15)
            r, g, b, a = random(), random(), random(), uniform(0.2, 0.8)

            # Emit a quad
            for index, (x, y) in enumerate(quad):
                geometry.append(x * scale + offset_x)
                geometry.append(y * scale + offset_y)
                # Color needs to be specified per vertex
                geometry.append(r)
                geometry.append(g)
                geometry.append(b)
                geometry.append(a)
                # Add to index buffer
                # NOTE: This is a rather silly index buffer not reusing indices
                index_buffer.append(index)

            # Create a rendering command for array rendering (no index buffer)
            # Expected data structure:
            # typedef  struct {
            #     uint  count;
            #     uint  instanceCount;
            #     uint  first;
            #     uint  baseInstance;
            # } DrawArraysIndirectCommand;
            array_commands.extend(
                (
                    6,  # count: How many vertices to render
                    1,  # instanceCount: How many instances. Typically 1.
                    quad_index * 6,  # first: What vertex index the object starts on
                    0,  # baseInstance: The base instance id (ignore when instanceCount = 1)
                )
            )
            # Create a rendering command for indexed rendering (with index buffer)
            # Expected data structure:
            # typedef  struct {
            #     GLuint  count;
            #     GLuint  instanceCount;
            #     GLuint  firstIndex;
            #     GLuint  baseVertex;
            #     GLuint  baseInstance;
            # } DrawElementsIndirectCommand;
            indexed_commands.extend(
                (
                    6,  # count: How many indices to render
                    1,  # instanceCount: How many instances to render. Typically 1
                    quad_index * 6,  # firstIndex: Position of the first index (ibo)
                    quad_index * 6,  # baseVertex: Position of the first vertex (vbo)
                    0,  # baseInstance: What value gl_InstanceID should start on
                )
            )

        return geometry, index_buffer, array_commands, indexed_commands

    def on_draw(self):
        self.clear()
        method = ""

        self.ctx.enable(self.ctx.BLEND)

        # Cycle all three rendering methods:
        # * Direct rendering of all quads
        # * Indirectly draw each individual quad (no index buffer)
        # * Indirectly draw each individual quad (with index buffer)
        with self.query:
            if self.frame % 3 == 0:
                # Draw the rectangles as one object
                # NOTE: first and vertices are not necessary if we are drawing all the geometry.
                method = "direct           "
                self.geometry_array.render(
                    self.program,
                    first=self.start * 6,
                    vertices=self.count * 6,
                )
            elif self.frame % 3 == 1:
                # Draw the rectangles as separate objects
                # NOTE: first and vertices are not necessary if we are drawing all the geometry.
                method = "indirect(array)  "
                self.geometry_array.render_indirect(
                    self.program,
                    self.draw_command_array_buffer,
                    first=self.start,
                    count=self.count,
                )
            else:
                method = "indirect(indexed)"
                # Draw rectangles as separate objects with index buffer
                # NOTE: first and vertices are not necessary if we are drawing all the geometry.
                self.geometry_indexed.render_indirect(
                    self.program,
                    self.draw_command_indexed_buffer,
                    first=self.start,
                    count=self.count,
                )

        # We want to prove by using a query that the following
        # is identical for all the rendering methods:
        # * The number of pixels written (rgba / 4)
        # * The number of primitives generate
        # We also compare the nanoseconds each methods is using
        # In addition we can tweak the start and count value
        # to prove that partial rendering also works.
        # NOTE: These values can be skewed if vsync is enabled
        print(

                f"[{method}] "
                f"Pixels written = {self.query.samples_passed // 4}, "
                f"Primitives Generated = {self.query.primitives_generated}, "
                f"time = {self.query.time_elapsed / 1_000_000_000}s"

        )

    def on_update(self, delta_time: float):
        self.frame += 1


RenderIndirect().run()



"""
Sprite Collect Moving and Bouncing Coins

Simple program to show basic sprite usage.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_collect_coins_move_bouncing
"""

import random
import arcade

# --- Constants ---
SPRITE_SCALING_PLAYER = 0.75
SPRITE_SCALING_COIN = 0.3
COIN_COUNT = 50

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Collect Moving and Bouncing Coins Example"


class Coin(arcade.Sprite):

    def __init__(self, filename, scale):

        super().__init__(filename, scale=scale)

        self.change_x = 0
        self.change_y = 0

    def update(self, delta_time: float = 1 / 60):

        # Move the coin
        self.center_x += self.change_x
        self.center_y += self.change_y

        # If we are out-of-bounds, then 'bounce'
        if self.left < 0:
            self.change_x *= -1

        if self.right > WINDOW_WIDTH:
            self.change_x *= -1

        if self.bottom < 0:
            self.change_y *= -1

        if self.top > WINDOW_HEIGHT:
            self.change_y *= -1


class GameView(arcade.View):

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.all_sprites_list = None
        self.coin_list = None

        # Set up the player info
        self.player_sprite = None
        self.score = 0

        # Don't show the mouse cursor
        self.window.set_mouse_visible(False)

        self.background_color = arcade.color.AMAZON

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.all_sprites_list = arcade.SpriteList()
        self.coin_list = arcade.SpriteList()

        # Score
        self.score = 0

        # Set up the player
        # Character image from kenney.nl
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING_PLAYER,
        )
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 50
        self.all_sprites_list.append(self.player_sprite)

        # Create the coins
        for i in range(50):

            # Create the coin instance
            # Coin image from kenney.nl
            coin = Coin(":resources:images/items/coinGold.png", scale=SPRITE_SCALING_COIN)

            # Position the coin
            coin.center_x = random.randrange(WINDOW_WIDTH)
            coin.center_y = random.randrange(WINDOW_HEIGHT)
            coin.change_x = random.randrange(-3, 4)
            coin.change_y = random.randrange(-3, 4)

            # Add the coin to the lists
            self.all_sprites_list.append(coin)
            self.coin_list.append(coin)

    def on_draw(self):
        """ Draw everything """
        self.clear()
        self.all_sprites_list.draw()

        # Put the text on the screen.
        output = f"Score: {self.score}"
        arcade.draw_text(output, 10, 20, arcade.color.WHITE, 14)

    def on_mouse_motion(self, x, y, dx, dy):
        """ Handle Mouse Motion """

        # Move the center of the player sprite to match the mouse x, y
        self.player_sprite.center_x = x
        self.player_sprite.center_y = y

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.all_sprites_list.update(delta_time)

        # Generate a list of all sprites that collided with the player.
        hit_list = arcade.check_for_collision_with_list(self.player_sprite,
                                                        self.coin_list)

        # Loop through each colliding sprite, remove it, and add to the score.
        for coin in hit_list:
            coin.remove_from_sprite_lists()
            self.score += 1


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



import typing
import weakref
from ctypes import (
    POINTER,
    byref,
    c_buffer,
    c_char,
    c_char_p,
    c_int,
    cast,
    create_string_buffer,
    pointer,
)
from typing import TYPE_CHECKING, Any, Iterable

from pyglet import gl

from .exceptions import ShaderException
from .types import SHADER_TYPE_NAMES, AttribFormat, GLTypes, PyGLenum
from .uniform import Uniform, UniformBlock

if TYPE_CHECKING:
    from arcade.gl import Context


class Program:
    """
    Compiled and linked shader program.

    Currently supports

    - vertex shader
    - fragment shader
    - geometry shader
    - tessellation control shader
    - tessellation evaluation shader

    Transform feedback also supported when output attributes
    names are passed in the varyings parameter.

    The best way to create a program instance is through :py:meth:`arcade.gl.Context.program`

    Args:
        ctx:
            The context this program belongs to
        vertex_shader:
            Vertex shader source
        fragment_shader:
            Fragment shader source
        geometry_shader:
            Geometry shader source
        tess_control_shader:
            Tessellation control shader source
        tess_evaluation_shader:
            Tessellation evaluation shader source
        varyings:
            List of out attributes used in transform feedback.
        varyings_capture_mode:
            The capture mode for transforms.
            ``"interleaved"`` means all out attribute will be written to a single buffer.
            ``"separate"`` means each out attribute will be written separate buffers.
            Based on these settings the `transform()` method will accept a single
            buffer or a list of buffer.
    """

    __slots__ = (
        "_ctx",
        "_glo",
        "_uniforms",
        "_varyings",
        "_varyings_capture_mode",
        "_geometry_info",
        "_attributes",
        "attribute_key",
        "__weakref__",
    )

    _valid_capture_modes = ("interleaved", "separate")

    def __init__(
        self,
        ctx: "Context",
        *,
        vertex_shader: str,
        fragment_shader: str | None = None,
        geometry_shader: str | None = None,
        tess_control_shader: str | None = None,
        tess_evaluation_shader: str | None = None,
        varyings: list[str] | None = None,
        varyings_capture_mode: str = "interleaved",
    ):
        self._ctx = ctx
        self._glo = glo = gl.glCreateProgram()
        self._varyings = varyings or []
        self._varyings_capture_mode = varyings_capture_mode.strip().lower()
        self._geometry_info = (0, 0, 0)
        self._attributes = []  # type: list[AttribFormat]
        #: Internal cache key used with vertex arrays
        self.attribute_key = "INVALID"  # type: str
        self._uniforms: dict[str, Uniform | UniformBlock] = {}

        if self._varyings_capture_mode not in self._valid_capture_modes:
            raise ValueError(
                f"Invalid capture mode '{self._varyings_capture_mode}'. "
                f"Valid modes are: {self._valid_capture_modes}."
            )

        shaders: list[tuple[str, int]] = [(vertex_shader, gl.GL_VERTEX_SHADER)]
        if fragment_shader:
            shaders.append((fragment_shader, gl.GL_FRAGMENT_SHADER))
        if geometry_shader:
            shaders.append((geometry_shader, gl.GL_GEOMETRY_SHADER))
        if tess_control_shader:
            shaders.append((tess_control_shader, gl.GL_TESS_CONTROL_SHADER))
        if tess_evaluation_shader:
            shaders.append((tess_evaluation_shader, gl.GL_TESS_EVALUATION_SHADER))

        # Inject a dummy fragment shader on gles when doing transforms
        if self._ctx.gl_api == "gles" and not fragment_shader:
            dummy_frag_src = """
                #version 310 es
                precision mediump float;
                out vec4 fragColor;
                void main() { fragColor = vec4(1.0); }
            """
            shaders.append((dummy_frag_src, gl.GL_FRAGMENT_SHADER))

        shaders_id = []
        for shader_code, shader_type in shaders:
            shader = Program.compile_shader(shader_code, shader_type)
            gl.glAttachShader(self._glo, shader)
            shaders_id.append(shader)

        # For now we assume varyings can be set up if no fragment shader
        if not fragment_shader:
            self._configure_varyings()

        Program.link(self._glo)

        if geometry_shader:
            geometry_in = gl.GLint()
            geometry_out = gl.GLint()
            geometry_vertices = gl.GLint()
            gl.glGetProgramiv(self._glo, gl.GL_GEOMETRY_INPUT_TYPE, geometry_in)
            gl.glGetProgramiv(self._glo, gl.GL_GEOMETRY_OUTPUT_TYPE, geometry_out)
            gl.glGetProgramiv(self._glo, gl.GL_GEOMETRY_VERTICES_OUT, geometry_vertices)
            self._geometry_info = (
                geometry_in.value,
                geometry_out.value,
                geometry_vertices.value,
            )

        # Delete shaders (not needed after linking)
        for shader in shaders_id:
            gl.glDeleteShader(shader)
            gl.glDetachShader(self._glo, shader)

        # Handle uniforms
        self._introspect_attributes()
        self._introspect_uniforms()
        self._introspect_uniform_blocks()

        if self._ctx.gc_mode == "auto":
            weakref.finalize(self, Program.delete_glo, self._ctx, glo)

        self.ctx.stats.incr("program")

    def __del__(self):
        # Intercept garbage collection if we are using Context.gc()
        if self._ctx.gc_mode == "context_gc" and self._glo > 0:
            self._ctx.objects.append(self)

    @property
    def ctx(self) -> "Context":
        """The context this program belongs to."""
        return self._ctx

    @property
    def glo(self) -> int:
        """The OpenGL resource id for this program."""
        return self._glo

    @property
    def attributes(self) -> Iterable[AttribFormat]:
        """List of attribute information."""
        return self._attributes

    @property
    def varyings(self) -> list[str]:
        """Out attributes names used in transform feedback."""
        return self._varyings

    @property
    def out_attributes(self) -> list[str]:
        """
        Out attributes names used in transform feedback.

        Alias for `varyings`.
        """
        return self._varyings

    @property
    def varyings_capture_mode(self) -> str:
        """
        Get the capture more for transform feedback (single, multiple).

        This is a read only property since capture mode
        can only be set before the program is linked.
        """
        return self._varyings_capture_mode

    @property
    def geometry_input(self) -> int:
        """
        The geometry shader's input primitive type.

        This an be compared with ``GL_TRIANGLES``, ``GL_POINTS`` etc.
        and is queried when the program is created.
        """
        return self._geometry_info[0]

    @property
    def geometry_output(self) -> int:
        """The geometry shader's output primitive type.

        This an be compared with ``GL_TRIANGLES``, ``GL_POINTS`` etc.
        and is queried when the program is created.
        """
        return self._geometry_info[1]

    @property
    def geometry_vertices(self) -> int:
        """
        The maximum number of vertices that can be emitted.
        This is queried when the program is created.
        """
        return self._geometry_info[2]

    def delete(self):
        """
        Destroy the underlying OpenGL resource.

        Don't use this unless you know exactly what you are doing.
        """
        Program.delete_glo(self._ctx, self._glo)
        self._glo = 0

    @staticmethod
    def delete_glo(ctx, prog_id):
        """
        Deletes a program. This is normally called automatically when the
        program is garbage collected.

        Args:
            ctx:
                The context this program belongs to
            prog_id:
                The OpenGL resource id
        """
        # Check to see if the context was already cleaned up from program
        # shut down. If so, we don't need to delete the shaders.
        if gl.current_context is None:
            return

        gl.glDeleteProgram(prog_id)
        ctx.stats.decr("program")

    def __getitem__(self, item) -> Uniform | UniformBlock:
        """Get a uniform or uniform block"""
        try:
            uniform = self._uniforms[item]
        except KeyError:
            raise KeyError(f"Uniform with the name `{item}` was not found.")

        return uniform.getter()

    def __setitem__(self, key, value):
        """
        Set a uniform value.

        Example::

            program['color'] = 1.0, 1.0, 1.0, 1.0
            program['mvp'] = projection @ view @ model

        Args:
            key:
                The uniform name
            value:
                The uniform value
        """
        try:
            uniform = self._uniforms[key]
        except KeyError:
            raise KeyError(f"Uniform with the name `{key}` was not found.")

        uniform.setter(value)

    def set_uniform_safe(self, name: str, value: Any):
        """
        Safely set a uniform catching KeyError.

        Args:
            name:
                The uniform name
            value:
                The uniform value
        """
        try:
            self[name] = value
        except KeyError:
            pass

    def set_uniform_array_safe(self, name: str, value: list[Any]):
        """
        Safely set a uniform array.

        Arrays can be shortened by the glsl compiler not all elements are determined
        to be in use. This function checks the length of the actual array and sets a
        subset of the values if needed. If the uniform don't exist no action will be
        done.

        Args:
            name:
                Name of uniform
            value:
                List of values
        """
        if name not in self._uniforms:
            return

        uniform = typing.cast(Uniform, self._uniforms[name])
        _len = uniform._array_length * uniform._components
        if _len == 1:
            self.set_uniform_safe(name, value[0])
        else:
            self.set_uniform_safe(name, value[:_len])

    def use(self):
        """
        Activates the shader.

        This is normally done for you automatically.
        """
        # IMPORTANT: This is the only place glUseProgram should be called
        #            so we can track active program.
        # if self._ctx.active_program != self:
        gl.glUseProgram(self._glo)
        # self._ctx.active_program = self

    def _configure_varyings(self):
        """Set up transform feedback varyings"""
        if not self._varyings:
            return

        # Covert names to char**
        c_array = (c_char_p * len(self._varyings))()
        for i, name in enumerate(self._varyings):
            c_array[i] = name.encode()

        ptr = cast(c_array, POINTER(POINTER(c_char)))

        # Are we capturing in interlaved or separate buffers?
        mode = (
            gl.GL_INTERLEAVED_ATTRIBS
            if self._varyings_capture_mode == "interleaved"
            else gl.GL_SEPARATE_ATTRIBS
        )

        gl.glTransformFeedbackVaryings(
            self._glo,  # program
            len(self._varyings),  # number of varying variables used for transform feedback
            ptr,  # zero-terminated strings specifying the names of the varying variables
            mode,
        )

    def _introspect_attributes(self):
        """Introspect and store detailed info about an attribute"""
        # TODO: Ensure gl_* attributes are ignored
        num_attrs = gl.GLint()
        gl.glGetProgramiv(self._glo, gl.GL_ACTIVE_ATTRIBUTES, num_attrs)
        num_varyings = gl.GLint()
        gl.glGetProgramiv(self._glo, gl.GL_TRANSFORM_FEEDBACK_VARYINGS, num_varyings)
        # print(f"attrs {num_attrs.value} varyings={num_varyings.value}")

        for i in range(num_attrs.value):
            c_name = create_string_buffer(256)
            c_size = gl.GLint()
            c_type = gl.GLenum()
            gl.glGetActiveAttrib(
                self._glo,  # program to query
                i,  # index (not the same as location)
                256,  # max attr name size
                None,  # c_length,  # length of name
                c_size,  # size of attribute (array or not)
                c_type,  # attribute type (enum)
                c_name,  # name buffer
            )

            # Get the actual location. Do not trust the original order
            location = gl.glGetAttribLocation(self._glo, c_name)

            # print(c_name.value, c_size, c_type)
            type_info = GLTypes.get(c_type.value)
            # print(type_info)
            self._attributes.append(
                AttribFormat(
                    c_name.value.decode(),
                    type_info.gl_type,
                    type_info.components,
                    type_info.gl_size,
                    location=location,
                )
            )

        # The attribute key is used to cache VertexArrays
        self.attribute_key = ":".join(
            f"{attr.name}[{attr.gl_type}/{attr.components}]" for attr in self._attributes
        )

    def _introspect_uniforms(self):
        """Figure out what uniforms are available and build an internal map"""
        # Number of active uniforms in the program
        active_uniforms = gl.GLint(0)
        gl.glGetProgramiv(self._glo, gl.GL_ACTIVE_UNIFORMS, byref(active_uniforms))

        # Loop all the active uniforms
        for index in range(active_uniforms.value):
            # Query uniform information like name, type, size etc.
            u_name, u_type, u_size = self._query_uniform(index)
            u_location = gl.glGetUniformLocation(self._glo, u_name.encode())

            # Skip uniforms that may be in Uniform Blocks
            # TODO: We should handle all uniforms
            if u_location == -1:
                # print(f"Uniform {u_location} {u_name} {u_size} {u_type} skipped")
                continue

            u_name = u_name.replace("[0]", "")  # Remove array suffix
            self._uniforms[u_name] = Uniform(
                self._ctx, self._glo, u_location, u_name, u_type, u_size
            )

    def _introspect_uniform_blocks(self):
        active_uniform_blocks = gl.GLint(0)
        gl.glGetProgramiv(self._glo, gl.GL_ACTIVE_UNIFORM_BLOCKS, byref(active_uniform_blocks))
        # print('GL_ACTIVE_UNIFORM_BLOCKS', active_uniform_blocks)

        for loc in range(active_uniform_blocks.value):
            index, size, name = self._query_uniform_block(loc)
            block = UniformBlock(self._glo, index, size, name)
            self._uniforms[name] = block

    def _query_uniform(self, location: int) -> tuple[str, int, int]:
        """Retrieve Uniform information at given location.

        Returns the name, the type as a GLenum (GL_FLOAT, ...) and the size. Size is
        greater than 1 only for Uniform arrays, like an array of floats or an array
        of Matrices.
        """
        u_size = gl.GLint()
        u_type = gl.GLenum()
        buf_size = 192  # max uniform character length
        u_name = create_string_buffer(buf_size)
        gl.glGetActiveUniform(
            self._glo,  # program to query
            location,  # location to query
            buf_size,  # size of the character/name buffer
            None,  # the number of characters actually written by OpenGL in the string
            u_size,  # size of the uniform variable
            u_type,  # data type of the uniform variable
            u_name,  # string buffer for storing the name
        )
        return u_name.value.decode(), u_type.value, u_size.value

    def _query_uniform_block(self, location: int) -> tuple[int, int, str]:
        """Query active uniform block by retrieving the name and index and size"""
        # Query name
        u_size = gl.GLint()
        buf_size = 192  # max uniform character length
        u_name = create_string_buffer(buf_size)
        gl.glGetActiveUniformBlockName(
            self._glo,  # program to query
            location,  # location to query
            256,  # max size if the name
            u_size,  # length
            u_name,
        )
        # Query index
        index = gl.glGetUniformBlockIndex(self._glo, u_name)
        # Query size
        b_size = gl.GLint()
        gl.glGetActiveUniformBlockiv(self._glo, index, gl.GL_UNIFORM_BLOCK_DATA_SIZE, b_size)
        return index, b_size.value, u_name.value.decode()

    @staticmethod
    def compile_shader(source: str, shader_type: PyGLenum) -> gl.GLuint:
        """
        Compile the shader code of the given type.

        Args:
            source:
                The shader source code
            shader_type:
                The type of shader to compile.
                ``GL_VERTEX_SHADER``, ``GL_FRAGMENT_SHADER`` etc.

        Returns:
            The created shader id
        """
        shader = gl.glCreateShader(shader_type)
        source_bytes = source.encode("utf-8")
        # Turn the source code string into an array of c_char_p arrays.
        strings = byref(cast(c_char_p(source_bytes), POINTER(c_char)))
        # Make an array with the strings lengths
        lengths = pointer(c_int(len(source_bytes)))
        gl.glShaderSource(shader, 1, strings, lengths)
        gl.glCompileShader(shader)
        result = c_int()
        gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS, byref(result))
        if result.value == gl.GL_FALSE:
            msg = create_string_buffer(512)
            length = c_int()
            gl.glGetShaderInfoLog(shader, 512, byref(length), msg)
            raise ShaderException(
                (
                    f"Error compiling {SHADER_TYPE_NAMES[shader_type]} "
                    f"({result.value}): {msg.value.decode('utf-8')}\n"
                    f"---- [{SHADER_TYPE_NAMES[shader_type]}] ---\n"
                )
                + "\n".join(
                    f"{str(i + 1).zfill(3)}: {line} " for i, line in enumerate(source.split("\n"))
                )
            )
        return shader

    @staticmethod
    def link(glo):
        """Link a shader program"""
        gl.glLinkProgram(glo)
        status = c_int()
        gl.glGetProgramiv(glo, gl.GL_LINK_STATUS, status)
        if not status.value:
            length = c_int()
            gl.glGetProgramiv(glo, gl.GL_INFO_LOG_LENGTH, length)
            log = c_buffer(length.value)
            gl.glGetProgramInfoLog(glo, len(log), None, log)
            raise ShaderException("Program link error: {}".format(log.value.decode()))

    def __repr__(self):
        return "<Program id={}>".format(self._glo)



"""
Create a maze using a depth-first search maze generation algorithm.
For more information on this algorithm see:
https://www.algosome.com/articles/maze-generation-depth-first.html
...or search up some other examples.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.maze_depth_first
"""
import random
import arcade
import timeit

NATIVE_SPRITE_SIZE = 128
SPRITE_SCALING = 0.25
SPRITE_SIZE = int(NATIVE_SPRITE_SIZE * SPRITE_SCALING)

WINDOW_WIDTH = 1000
WINDOW_HEIGHT = 700
WINDOW_TITLE = "Maze Depth First Example"

MOVEMENT_SPEED = 8

TILE_EMPTY = 0
TILE_CRATE = 1

# Maze must have an ODD number of rows and columns.
# Walls go on EVEN rows/columns.
# Openings go on ODD rows/columns
MAZE_HEIGHT = 51
MAZE_WIDTH = 51

MERGE_SPRITES = True

# How many pixels to keep as a minimum margin between the character
# and the edge of the screen.
VIEWPORT_MARGIN = 200
HORIZONTAL_BOUNDARY = WINDOW_WIDTH / 2.0 - VIEWPORT_MARGIN
VERTICAL_BOUNDARY = WINDOW_HEIGHT / 2.0 - VIEWPORT_MARGIN
# If the player moves further than this boundary away from
# he camera we use a constraint to move the camera
CAMERA_BOUNDARY = arcade.LRBT(
    -HORIZONTAL_BOUNDARY,
    HORIZONTAL_BOUNDARY,
    -VERTICAL_BOUNDARY,
    VERTICAL_BOUNDARY,
)

def _create_grid_with_cells(width, height):
    """ Create a grid with empty cells on odd row/column combinations. """
    grid = []
    for row in range(height):
        grid.append([])
        for column in range(width):
            if column % 2 == 1 and row % 2 == 1:
                grid[row].append(TILE_EMPTY)
            elif column == 0 or row == 0 or column == width - 1 or row == height - 1:
                grid[row].append(TILE_CRATE)
            else:
                grid[row].append(TILE_CRATE)
    return grid


def make_maze_depth_first(maze_width, maze_height):
    maze = _create_grid_with_cells(maze_width, maze_height)

    w = (len(maze[0]) - 1) // 2
    h = (len(maze) - 1) // 2
    vis = [[0] * w + [1] for _ in range(h)] + [[1] * (w + 1)]

    def walk(x: int, y: int):
        vis[y][x] = 1

        d = [(x - 1, y), (x, y + 1), (x + 1, y), (x, y - 1)]
        random.shuffle(d)
        for (xx, yy) in d:
            if vis[yy][xx]:
                continue
            if xx == x:
                maze[max(y, yy) * 2][x * 2 + 1] = TILE_EMPTY
            if yy == y:
                maze[y * 2 + 1][max(x, xx) * 2] = TILE_EMPTY

            walk(xx, yy)

    walk(random.randrange(w), random.randrange(h))

    return maze


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """
        Initializer
        """
        super().__init__()

        # Sprite lists
        self.player_list = None
        self.wall_list = None

        # Player info
        self.score = 0
        self.player_sprite = None

        # Physics engine
        self.physics_engine = None

        # Camera for scrolling
        self.camera = None

        # Time to process
        self.processing_time = 0
        self.draw_time = 0

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()
        self.wall_list = arcade.SpriteList()

        self.score = 0

        # Create the maze
        maze = make_maze_depth_first(MAZE_WIDTH, MAZE_HEIGHT)

        # Create sprites based on 2D grid
        if not MERGE_SPRITES:
            # This is the simple-to-understand method. Each grid location
            # is a sprite.
            for row in range(MAZE_HEIGHT):
                for column in range(MAZE_WIDTH):
                    if maze[row][column] == 1:
                        wall = arcade.Sprite(
                            ":resources:images/tiles/grassCenter.png",
                            scale=SPRITE_SCALING,
                        )
                        wall.center_x = column * SPRITE_SIZE + SPRITE_SIZE / 2
                        wall.center_y = row * SPRITE_SIZE + SPRITE_SIZE / 2
                        self.wall_list.append(wall)
        else:
            # This uses new Arcade 1.3.1 features, that allow me to create a
            # larger sprite with a repeating texture. So if there are multiple
            # cells in a row with a wall, we merge them into one sprite, with a
            # repeating texture for each cell. This reduces our sprite count.
            for row in range(MAZE_HEIGHT):
                column = 0
                while column < len(maze):
                    while column < len(maze) and maze[row][column] == 0:
                        column += 1
                    start_column = column
                    while column < len(maze) and maze[row][column] == 1:
                        column += 1
                    end_column = column - 1

                    column_count = end_column - start_column + 1
                    column_mid = (start_column + end_column) / 2

                    wall = arcade.Sprite(
                        ":resources:images/tiles/grassCenter.png",
                        scale=SPRITE_SCALING,
                    )
                    wall.center_x = column_mid * SPRITE_SIZE + SPRITE_SIZE / 2
                    wall.center_y = row * SPRITE_SIZE + SPRITE_SIZE / 2
                    wall.width = SPRITE_SIZE * column_count
                    self.wall_list.append(wall)

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING)
        self.player_list.append(self.player_sprite)

        # Randomly place the player. If we are in a wall, repeat until we aren't.
        placed = False
        while not placed:

            # Randomly position
            self.player_sprite.center_x = random.randrange(MAZE_WIDTH * SPRITE_SIZE)
            self.player_sprite.center_y = random.randrange(MAZE_HEIGHT * SPRITE_SIZE)

            # Are we in a wall?
            walls_hit = arcade.check_for_collision_with_list(self.player_sprite, self.wall_list)
            if len(walls_hit) == 0:
                # Not in a wall! Success!
                placed = True

        self.physics_engine = arcade.PhysicsEngineSimple(self.player_sprite, self.wall_list)

        # Set the background color
        self.background_color = arcade.color.AMAZON

        # Setup Camera
        self.camera = arcade.Camera2D()

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Start timing how long this takes
        draw_start_time = timeit.default_timer()

        # Draw all the sprites.
        self.wall_list.draw()
        self.player_list.draw()

        # Draw info on the screen
        sprite_count = len(self.wall_list)

        output = f"Sprite Count: {sprite_count}"
        left, bottom = self.camera.bottom_left
        arcade.draw_text(output,
                         left + 20,
                         WINDOW_HEIGHT - 20 + bottom,
                         arcade.color.WHITE, 16)

        output = f"Drawing time: {self.draw_time:.3f}"
        arcade.draw_text(output,
                         left + 20,
                         WINDOW_HEIGHT - 40 + bottom,
                         arcade.color.WHITE, 16)

        output = f"Processing time: {self.processing_time:.3f}"
        arcade.draw_text(output,
                         left + 20,
                         WINDOW_HEIGHT - 60 + bottom,
                         arcade.color.WHITE, 16)

        self.draw_time = timeit.default_timer() - draw_start_time

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.player_sprite.change_y = MOVEMENT_SPEED
        elif key == arcade.key.DOWN:
            self.player_sprite.change_y = -MOVEMENT_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP or key == arcade.key.DOWN:
            self.player_sprite.change_y = 0
        elif key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0

    def on_update(self, delta_time):
        """ Movement and game logic """

        start_time = timeit.default_timer()

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.physics_engine.update()

        # --- Manage Scrolling ---
        self.camera.position = arcade.camera.grips.constrain_boundary_xy(
            self.camera.view_data, CAMERA_BOUNDARY, self.player_sprite.position
        )
        self.camera.use()

        # Save the time it took to do this.
        self.processing_time = timeit.default_timer() - start_time


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()



if __name__ == "__main__":
    main()



from __future__ import annotations

from abc import ABC
from collections.abc import Iterable
from enum import IntEnum
from typing import TYPE_CHECKING, NamedTuple, TypeVar

from pyglet.event import EVENT_HANDLED, EVENT_UNHANDLED, EventDispatcher
from pyglet.math import Vec2
from typing_extensions import Self

import arcade
from arcade import LBWH, Rect, Sprite, Texture
from arcade.color import TRANSPARENT_BLACK
from arcade.gui.events import (
    UIEvent,
    UIMouseMovementEvent,
    UIMousePressEvent,
    UIMouseReleaseEvent,
    UIOnClickEvent,
    UIOnUpdateEvent,
)
from arcade.gui.nine_patch import NinePatchTexture
from arcade.gui.property import ListProperty, Property, bind
from arcade.gui.surface import Surface
from arcade.types import AnchorPoint, AsFloat, Color
from arcade.utils import copy_dunders_unimplemented

if TYPE_CHECKING:
    from arcade.gui.ui_manager import UIManager

W = TypeVar("W", bound="UIWidget")


class FocusMode(IntEnum):
    """Defines the focus mode of a widget.

    0: Not focusable
    1: Focusable

    We might support different focus modes in the future, but for now on/off is enough.
    """

    NONE = 0
    ALL = 2


class _ChildEntry(NamedTuple):
    child: UIWidget
    data: dict


@copy_dunders_unimplemented
class UIWidget(EventDispatcher, ABC):
    """The :class:`UIWidget` class is the base class required for creating widgets.

    We also have some default values and behaviors that you should be aware of:

    * A :class:`UIWidget` is not a :class:`~arcade.gui.UILayout`: it will not
      change the position or the size of its children. If you want control over
      positioning or sizing, use a :class:`~arcade.gui.UILayout`.

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: width of widget
        height: height of widget
        size_hint: Tuple of floats (0.0-1.0), how much space of the parent should be requested
        size_hint_min: min width and height in pixel
        size_hint_max: max width and height in pixel
    """

    rect = Property(LBWH(0, 0, 1, 1))
    visible = Property(True)
    focused = Property(False)
    focus_mode: FocusMode = FocusMode.NONE

    size_hint = Property[tuple[float | None, float | None] | None](None)
    size_hint_min = Property[tuple[float | None, float | None] | None](None)
    size_hint_max = Property[tuple[float | None, float | None] | None](None)

    _children = ListProperty[_ChildEntry]()
    _border_width = Property(0)
    _border_color = Property[Color | None](arcade.color.BLACK)
    _bg_color = Property[Color | None]()
    _bg_tex = Property[Texture | NinePatchTexture | None]()
    _padding_top = Property(0)
    _padding_right = Property(0)
    _padding_bottom = Property(0)
    _padding_left = Property(0)
    _strong_background = Property(False)
    """If True, the background will clear the surface, even if it is not fully opaque.
    This is not part of the public API and subject to change.
    UILabel have a strong background if set.
    """

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float = 100,
        height: float = 100,
        children: Iterable[UIWidget] = tuple(),
        # Properties which might be used by layouts
        size_hint: tuple[float | None, float | None] | None = None,  # in percentage
        size_hint_min: tuple[float | None, float | None] | None = None,  # in pixel
        size_hint_max: tuple[float | None, float | None] | None = None,  # in pixel
        **kwargs,
    ):
        self._requires_render = True
        self.rect = LBWH(x, y, width, height)
        self.parent: UIManager | UIWidget | None = None

        # Size hints are properties that can be used by layouts
        self.size_hint = size_hint
        self.size_hint_min = size_hint_min
        self.size_hint_max = size_hint_max

        self.register_event_type("on_event")
        self.register_event_type("on_update")

        for child in children:
            self.add(child)

        bind(self, "rect", self.trigger_full_render)
        bind(self, "focused", self.trigger_full_render)
        bind(
            self, "visible", self.trigger_full_render
        )  # TODO maybe trigger_parent_render would be enough
        bind(self, "_children", self.trigger_render)
        bind(self, "_border_width", self.trigger_render)
        bind(self, "_border_color", self.trigger_render)
        bind(self, "_bg_color", self.trigger_render)
        bind(self, "_bg_tex", self.trigger_render)
        bind(self, "_padding_top", self.trigger_render)
        bind(self, "_padding_right", self.trigger_render)
        bind(self, "_padding_bottom", self.trigger_render)
        bind(self, "_padding_left", self.trigger_render)
        bind(self, "_strong_background", self.trigger_render)

    def add(self, child: W, **kwargs) -> W:
        """Add a widget as a child.

        Added widgets will receive UI events and be rendered.

        By default, the latest added widget will receive ui events first and will
        be rendered on top of others.

        Args:
            child: widget to add
            index: position a widget is added, None has the highest
                priority

        Returns:
            given child
        """
        child.parent = self
        index = kwargs.pop("index") if "index" in kwargs else None
        if index is None:
            self._children.append(_ChildEntry(child, kwargs))
        else:
            if not 0 <= index <= len(self.children):
                raise ValueError("Index must be between 0 and the number of children")
            self._children.insert(index, _ChildEntry(child, kwargs))

        return child

    def remove(self, child: UIWidget) -> dict | None:
        """Removes a child from the UIManager which was directly added to it.
        This will not remove widgets which are added to a child of UIManager.

        Return:
            kwargs which were used when child was added
        """
        child.parent = None
        for c in self._children:
            if c.child == child:
                self._children.remove(c)
                return c.data
        return None

    def clear(self):
        """Removes all children"""
        for child in self.children:
            child.parent = None

        self._children.clear()

    def __contains__(self, item):
        return item in self.children

    def on_update(self, dt):
        """Custom logic which will be triggered."""
        pass

    def on_event(self, event: UIEvent) -> bool | None:
        """Passes :class:`UIEvent` s through the widget tree."""
        # UpdateEvents are past to the first invisible widget
        if isinstance(event, UIOnUpdateEvent):
            self.dispatch_event("on_update", event.dt)

        if self.visible:
            # pass event to children
            for child in reversed(self.children):
                if child.dispatch_event("on_event", event):
                    return EVENT_HANDLED

        return EVENT_UNHANDLED

    def _walk_parents(self) -> Iterable[UIWidget | UIManager]:
        parent = self.parent
        while isinstance(parent, UIWidget):
            yield parent
            parent = parent.parent

        if parent:
            yield parent

    def trigger_render(self):
        """This will delay a render right before the next frame is rendered, so that
        :meth:`UIWidget.do_render` is not called multiple times.
        """
        self._requires_render = True

    def trigger_full_render(self) -> None:
        """In case a widget uses transparent areas or was moved,
        it might be important to request a full rendering of parents
        """
        self.trigger_render()
        for parent in self._walk_parents():
            parent.trigger_render()

    def _prepare_layout(self):
        """Helper function to trigger :meth:`UILayout.prepare_layout` through the widget tree,
        should only be used internally!
        """
        for child in self.children:
            child._prepare_layout()

    def _do_layout(self):
        """Helper function to trigger :meth:`UIWidget.do_layout` through the widget tree,
        should only be used by UIManager!
        """
        for child in self.children:
            # rect changes in children will trigger_full_render
            child._do_layout()

    def _do_render(self, surface: Surface, force=False) -> bool:
        """Helper function to trigger :meth:`UIWidget.do_render` through the widget tree,
        should only be used by UIManager!

        Returns:
            if this widget or a child was rendered
        """
        rendered = False

        should_render = force or self._requires_render
        if should_render and self.visible:
            rendered = True
            self.do_render_base(surface)
            self.do_render(surface)
            if self.focused:
                self.do_render_focus(surface)
            self._requires_render = False

        # only render children if self is visible
        if self.visible:
            for child in self.children:
                rendered |= child._do_render(surface, should_render)

        return rendered

    def do_render_base(self, surface: Surface):
        """Renders background, border and "padding"""
        surface.limit(self.rect)

        # draw background
        if self._bg_color:
            if self._bg_color.a == 255 or self._strong_background:
                surface.clear(self._bg_color)
            else:
                arcade.draw_rect_filled(LBWH(0, 0, self.width, self.height), color=self._bg_color)
        # draw background texture
        if self._bg_tex:
            surface.draw_texture(x=0, y=0, width=self.width, height=self.height, tex=self._bg_tex)

        # draw border
        if self._border_width and self._border_color:
            arcade.draw_lbwh_rectangle_outline(
                0,
                0,
                self.width,
                self.height,
                color=self._border_color,
                border_width=self._border_width * 2,
            )

    def prepare_render(self, surface):
        """Helper for rendering, the drawing area will be adjusted to the own position and size.
        Draw calls have to be relative to 0,0.
        This will also prevent any overdraw outside of the widgets area

        Args:
            surface: Surface used for rendering
        """
        surface.limit(self.content_rect)

    def do_render(self, surface: Surface):
        """Render the widgets graphical representation, use :meth:`UIWidget.prepare_render`
        to limit the drawing area to the widgets rect and draw relative to 0,0.
        """
        pass

    def do_render_focus(self, surface: Surface):
        """Render the widgets focus representation overlay`"""
        self.prepare_render(surface)
        arcade.draw_rect_outline(
            rect=LBWH(0, 0, self.content_width, self.content_height),
            color=arcade.color.WHITE,
            border_width=4,
        )

    def dispatch_ui_event(self, event: UIEvent):
        """Dispatch a :class:`UIEvent` using pyglet event dispatch mechanism"""
        return self.dispatch_event("on_event", event)

    def move(self, dx=0, dy=0):
        """Move the widget by dx and dy.

        Args:
            dx: x axis difference
            dy: y axis difference
        """
        self.rect = self.rect.move(dx, dy)

    def scale(self, factor: AsFloat, anchor: Vec2 = AnchorPoint.CENTER):
        """Scales the size of the widget (x,y,width, height) by factor.

        Args:
            factor: scale factor
            anchor: anchor point
        """
        self.rect = self.rect.scale(new_scale=factor, anchor=anchor)

    def get_ui_manager(self) -> UIManager | None:
        """The UIManager this widget is attached to. During creation, this will be None."""
        from arcade.gui.ui_manager import UIManager

        w: UIWidget | None = self
        while w and w.parent:
            parent = w.parent
            if isinstance(parent, UIManager):
                return parent

            w = parent
        return None

    @property
    def left(self) -> float:
        """Left coordinate of the widget"""
        return self.rect.left

    @property
    def right(self) -> float:
        """Right coordinate of the widget"""
        return self.rect.right

    @property
    def bottom(self) -> float:
        """Bottom coordinate of the widget"""
        return self.rect.bottom

    @property
    def top(self) -> float:
        """Top coordinate of the widget"""
        return self.rect.top

    @property
    def position(self) -> Vec2:
        """Returns bottom left coordinates"""
        return self.rect.bottom_left

    @property
    def center(self) -> Vec2:
        """Returns center coordinates"""
        return self.rect.center

    @center.setter
    def center(self, value: tuple[int, int]):
        self.rect = self.rect.align_center(value)

    @property
    def center_x(self) -> float:
        """Center x coordinate"""
        return self.rect.x

    @property
    def center_y(self) -> float:
        """Center y coordinate"""
        return self.rect.y

    @property
    def padding(self):
        """Returns padding as tuple (top, right, bottom, left)"""
        return (
            self._padding_top,
            self._padding_right,
            self._padding_bottom,
            self._padding_left,
        )

    @padding.setter
    def padding(self, args: int | tuple[int, int] | tuple[int, int, int, int]):
        if isinstance(args, int):  # self.padding = 10 -> 10, 10, 10, 10
            args = (args, args, args, args)

        elif len(args) == 2:  # self.padding = 10, 20 -> 10, 20, 10, 20
            args = args + args

        pt, pr, pb, pl = args  # self.padding = 10, 20, 30, 40
        self._padding_top = pt
        self._padding_right = pr
        self._padding_bottom = pb
        self._padding_left = pl

    @property
    def children(self) -> list[UIWidget]:
        """Provides all child widgets."""
        return [child for child, data in self._children]

    def __iter__(self):
        return iter(self.children)

    def resize(self, *, width=None, height=None, anchor: Vec2 = AnchorPoint.CENTER):
        """Resizes the widget.

        Args:
            width: new width
            height: new height
            anchor: anchor point for resizing, default is center
        """
        self.rect = self.rect.resize(width=width, height=height, anchor=anchor)

    def with_border(self, *, width=2, color: Color | None = arcade.color.GRAY) -> Self:
        """Sets border properties

        Args:
            width: border width
            color: border color

        Returns:
            self
        """
        self._border_width = width
        self._border_color = color
        return self

    def with_padding(
        self,
        *,
        top: int | None = None,
        right: int | None = None,
        bottom: int | None = None,
        left: int | None = None,
        all: int | None = None,
    ) -> Self:
        """Changes the padding to the given values if set. Returns itself

        Returns:
            self
        """
        if all is not None:
            self.padding = all
        if top is not None:
            self._padding_top = top
        if right is not None:
            self._padding_right = right
        if bottom is not None:
            self._padding_bottom = bottom
        if left is not None:
            self._padding_left = left
        return self

    def with_background(
        self,
        *,
        color: None | Color = ...,  # type: ignore
        texture: None | Texture | NinePatchTexture = ...,  # type: ignore
    ) -> Self:
        """Set widgets background.

        A color or texture can be used for background,
        if a texture is given, start and end point can be added to use the texture as ninepatch.

        Args:
            color: A color used as background
            texture: A texture or ninepatch texture used as background

        Returns:
            self
        """
        if color is not ...:
            if color is not None:
                color = Color.from_iterable(color)

            self._bg_color = color

        if texture is not ...:
            self._bg_tex = texture

        return self

    @property
    def content_size(self) -> tuple[float, float]:
        """Returns the size of the content area,
        which is the size of the widget minus padding and border."""
        return self.content_width, self.content_height

    @property
    def content_width(self) -> float:
        """Returns the width of the content area,
        which is the width of the widget minus padding and border."""
        return self.rect.width - 2 * self._border_width - self._padding_left - self._padding_right

    @property
    def content_height(self) -> float:
        """Returns the height of the content area,
        which is the height of the widget minus padding and border."""
        return self.rect.height - 2 * self._border_width - self._padding_top - self._padding_bottom

    @property
    def content_rect(self) -> Rect:
        """Returns the content area as a rect.
        The content area is the area of the widget minus padding and border."""
        return LBWH(
            self.left + self._border_width + self._padding_left,
            self.bottom + self._border_width + self._padding_bottom,
            self.content_width,
            self.content_height,
        )

    @property
    def width(self) -> float:
        """Width of the widget."""
        return self.rect.width

    @property
    def height(self) -> float:
        """Height of the widget."""
        return self.rect.height

    @property
    def size(self) -> Vec2:
        """Size of the widget."""
        return Vec2(self.width, self.height)

    def center_on_screen(self: W) -> W:
        """Places this widget in the center of the current window.

        This is a convenience method for simple centering of widgets without using
        a layout.

        In general, it is recommended to use layouts for more complex UIs.
        """
        center = arcade.get_window().center
        self.rect = self.rect.align_center(center)
        return self

    def __str__(self):
        return f"{self.__class__.__name__}()"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.rect.lbwh}>"


class UIInteractiveWidget(UIWidget):
    """Base class for widgets which use mouse interaction (hover, pressed, clicked)

    It provides properties for hovered, pressed and disabled states.

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: width of widget
        height: height of widget
        size_hint: Tuple of floats (0.0-1.0), how much space of the
            parent should be requested
        size_hint_min: min width and height in pixel
        size_hint_max: max width and height in pixel
        interaction_buttons: defines, which mouse buttons should trigger
            the interaction (default: left mouse button)
    """

    focus_mode = FocusMode.ALL

    # States
    hovered = Property(False)
    """True if the mouse is over the widget"""
    pressed = Property(False)
    """True if the widget is pressed"""
    disabled = Property(False)
    """True if the widget is disabled"""

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float = 100,
        height: float = 100,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        interaction_buttons=(arcade.MOUSE_BUTTON_LEFT,),
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            **kwargs,
        )
        self.register_event_type("on_click")

        self.interaction_buttons = interaction_buttons

        bind(self, "pressed", self.trigger_render)
        bind(self, "hovered", self.trigger_render)
        bind(self, "disabled", self.trigger_render)

    def on_event(self, event: UIEvent) -> bool | None:
        """Handles mouse events and triggers on_click event if the widget is clicked.

        This also sets the hovered and pressed state of the widget.
        """
        if super().on_event(event):
            return EVENT_HANDLED

        if isinstance(event, UIMouseMovementEvent):
            self.hovered = self.rect.point_in_rect(event.pos)

        if (
            isinstance(event, UIMousePressEvent)
            and self.rect.point_in_rect(event.pos)
            and event.button in self.interaction_buttons
        ):
            self.pressed = True
            return EVENT_HANDLED

        if (
            self.pressed
            and isinstance(event, UIMouseReleaseEvent)
            and event.button in self.interaction_buttons
        ):
            self.pressed = False
            if self.rect.point_in_rect(event.pos):
                if not self.disabled:
                    # Dispatch new on_click event, source is this widget itself
                    self.dispatch_event(
                        "on_click",
                        UIOnClickEvent(
                            source=self,
                            x=event.x,
                            y=event.y,
                            button=event.button,
                            modifiers=event.modifiers,
                        ),
                    )
                    return EVENT_HANDLED

        return EVENT_UNHANDLED

    def on_click(self, event: UIOnClickEvent):
        """Triggered when the widget is clicked."""
        pass


class UIDummy(UIInteractiveWidget):
    """Solid color widget used for testing & examples.

    Starts with a random color.
    It should not be subclassed for real-world usage.

    When clicked, it does the following:

    * Outputs its `rect` attribute to the console
    * Changes its color to a random fully opaque color

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: width of widget
        height: height of widget
        size_hint: Tuple of floats (0.0-1.0), how much space of the
            parent should be requested
        size_hint_min: min width and height in pixel
        size_hint_max: max width and height in pixel
        **kwargs: passed to UIWidget
    """

    def __init__(
        self,
        *,
        x=0,
        y=0,
        width=100,
        height=100,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            **kwargs,
        )
        self.with_background(color=Color.random(a=255))
        self.with_border(color=arcade.color.BATTLESHIP_GREY, width=0)

    def on_click(self, event: UIOnClickEvent):
        """Prints the rect and changes the color"""
        print("UIDummy.rect:", self.rect)
        self.with_background(color=Color.random(a=255))

    def on_update(self, dt):
        """Update the border of the widget if hovered"""
        self.with_border(
            width=2 if self.hovered else 0,
            color=arcade.color.WHITE if self.pressed else arcade.color.BATTLESHIP_GREY,
        )


class UISpriteWidget(UIWidget):
    """Create a UI element with a sprite that controls what is displayed.

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: width of widget
        height: height of widget
        sprite: Sprite to embed in gui
        size_hint: Tuple of floats (0.0-1.0), how much space of the
            parent should be requested
        size_hint_min: min width and height in pixel
        size_hint_max: max width and height in pixel
    """

    def __init__(
        self,
        *,
        x=0,
        y=0,
        width=100,
        height=100,
        sprite: Sprite | None = None,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            **kwargs,
        )
        self._sprite = sprite

    def on_update(self, dt):
        """Pass update event to sprite"""
        if self._sprite:
            self._sprite.update()
            self._sprite.update_animation(dt)
            self.trigger_render()

    def do_render(self, surface: Surface):
        """Render the sprite"""
        self.prepare_render(surface)
        surface.clear(color=TRANSPARENT_BLACK)
        if self._sprite is not None:
            surface.draw_sprite(0, 0, self.width, self.height, self._sprite)


class UILayout(UIWidget):
    """Base class for widgets, which position themselves or their children.

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: width of widget
        height: height of widget
        children: Child widgets of this group
        size_hint: Tuple of floats (0.0-1.0), how much space of the
            parent should be requested
        size_hint_min: min width and height in pixel
        size_hint_max: max width and height in pixel
    """

    @staticmethod
    def min_size_of(child: UIWidget) -> tuple[float, float]:
        """Resolves the minimum size of a child. If it has a size_hint set for the axis,
        it will use size_hint_min if set, otherwise the actual size will be used.
        """
        sh_w, sh_h = child.size_hint or (None, None)
        shmn_w, shmn_h = child.size_hint_min or (None, None)

        min_w, min_h = child.size

        if sh_w is not None:
            min_w = shmn_w or 0

        if sh_h is not None:
            min_h = shmn_h or 0

        return min_w, min_h

    def _prepare_layout(self):
        """Triggered to prepare layout of this widget and its children.
        Common example is to update size_hints(min/max).
        """
        super()._prepare_layout()
        self.prepare_layout()

    def prepare_layout(self):
        """Triggered by the UIManager before layouting,
        :class:`UILayout` s should prepare themselves based on children.

        Prepare layout is triggered on children first.
        """
        pass

    def _do_layout(self):
        # rect change will trigger full render automatically
        self.do_layout()

        # Continue do_layout within subtree
        super()._do_layout()

    def do_layout(self):
        """do_layout is triggered by the UIManager before rendering.
         :class:`UILayout` should position their children.
         Afterward, do_layout of child widgets will be triggered.

        Use :meth:`UIWidget.trigger_render` to trigger a rendering before the next
        frame, this will happen automatically if the position or size of this widget changed.
        """


class UISpace(UIWidget):
    """Widget reserving space, can also have a background color.

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: width of widget
        height: height of widget
        color: Color for widget area, if None, it will be transparent
            (this will set the background color)
        size_hint: Tuple of floats (0.0-1.0), how much space of the
            parent should be requested
        size_hint_min: min width and height in pixel
        size_hint_max: max width and height in pixel
        **kwargs: passed to UIWidget
    """

    def __init__(
        self,
        *,
        x=0,
        y=0,
        width=1,
        height=1,
        color=None,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        **kwargs,
    ):
        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            **kwargs,
        )
        self.with_background(color=color)

    @property
    def color(self):
        """Color of the widget, alias for background color"""
        return self._bg_color

    @color.setter
    def color(self, value):
        self.with_background(color=value)


__all__ = ["Surface", "UIDummy", "FocusMode", "UIInteractiveWidget", "UIWidget"]
